/*
generalized class for obtaining accelerations
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during winter 2013
*/

#include "Acceleration.h"
#include <Arduino.h>			/// delay() for Atmel chips
#include <math.h>				/// to get M_PI


template <typename Driver> void Acceleration<Driver>::initialize(float xMin, float xMax, float yMin, float yMax, float zMin, float zMax)
{
  /// Set calibration data. These numbers has been found by collecting raw data during slowly varying movements. 9.81 m/s^2 is earths gravitational acceleration.
  _calibrationGainMin[0] = xMin;  /// Smallest measured raw value
  _calibrationGainMax[0] = xMax;  /// Highest value
  _calibrationGainMin[1] = yMin;
  _calibrationGainMax[1] = yMax;  
  _calibrationGainMin[2] = zMin;
  _calibrationGainMax[2] = zMax;
  
  if (_g == 0)
  {
		_g = 2;	/// default value
  }

  _driver.initialize();
}


/// input values must be raw! no conversion here
template <typename Driver> void Acceleration<Driver>::bias(float x, float y, float z)
{
	/// Calibration found once(!) by positioning the accelerometer perfectly level and record readings
  _bias[0] = x;  /// X
  _bias[1] = y;  /// Y
  _bias[2] = z;  /// Z
}


/// Read raw values from accelerometer
template <typename Driver> void Acceleration<Driver>::raw()
{
  int16_t * buffer = _driver.measure();
  
  /// FIFO stack
  _xAxis[2] = _xAxis[1];
  _xAxis[1] = _xAxis[0];
  _xAxis[0] = buffer[0];
  _yAxis[2] = _yAxis[1];
  _yAxis[1] = _yAxis[0];
  _yAxis[0] = buffer[1];
  _zAxis[2] = _zAxis[1];
  _zAxis[1] = _zAxis[0];
  _zAxis[0] = buffer[2];
}



template <typename Driver> void Acceleration<Driver>::measure()
{
  int16_t * buffer = _driver.measure();

  float x = buffer[0] - _bias[0];
  float y = buffer[1] - _bias[1];
  float z = buffer[2] - _bias[2];
  
  /// Convert to m/s^2. Max and min values for each axis is different, therefore different gain will be applied.
  x *= (x < 0 ? _calibrationGainMin[0] : _calibrationGainMax[0] );
  y *= (y < 0 ? _calibrationGainMin[1] : _calibrationGainMax[1] );
  z *= (z < 0 ? _calibrationGainMin[2] : _calibrationGainMax[2] );
  
  /// do not accept values outside the configured gravity scope
  float threshold = (float)_g * 9.81 * 1.1;
  if (abs(x) > threshold) x = _xAxis[0];
  if (abs(y) > threshold) y = _yAxis[0];
  if (abs(z) > threshold) z = _zAxis[0];
  
  /// FIFO stack
  _xAxis[2] = _xAxis[1];
  _xAxis[1] = _xAxis[0];
  _xAxis[0] = x;
  _yAxis[2] = _yAxis[1];
  _yAxis[1] = _yAxis[0];
  _yAxis[0] = y;
  _zAxis[2] = _zAxis[1];
  _zAxis[1] = _zAxis[0];
  _zAxis[0] = z;
}


/// Get acceleration in m/s^2 in x: axis=0, y: axis=1 or z: axis=2 directions
template <typename Driver> float Acceleration<Driver>::axis(int axis, int index)
{ 
  if (index > 2)
  {
    index = 0;
  }

  /// Values below are negative because of mounting direction of accelerometer
  if (axis == 1)
  {
    return _yAxis[index];
  }
  else if (axis == 2)
  {
    return _zAxis[index];
  }
  else
  {
    return _xAxis[index];
  }
}



template <typename Driver> float Acceleration<Driver>::x(int index)
{
  return axis(0,index);
}
template <typename Driver> float Acceleration<Driver>::y(int index)
{
  return axis(1,index);
}
template <typename Driver> float Acceleration<Driver>::z(int index)
{
  return axis(2,index);
}


/// set the desired gravitational force
template <typename Driver> bool Acceleration<Driver>::setG(int g)
{
	if (_driver.setG(g))
	{
		_g = g;
		return true;
	}
	return false;
}



template <typename Driver> bool Acceleration<Driver>::setHz(float hz)
{
	return _driver.setHz(hz);
}



template <typename Driver> float Acceleration<Driver>::roll()  /// phi
{
  float angle = atan2(y(),z()) - M_PI;
  if (angle < 0.0)
  {
    angle += M_PI*2;
  }
  return angle;
}



template <typename Driver> float Acceleration<Driver>::pitch()  /// theta
{
  float angle = M_PI - atan2(x(),z());
  if (angle < 0.0)
  {
    angle += M_PI*2;
  }
  return angle;
}



template <typename Driver> bool Acceleration<Driver>::calibrate(bool calibrateXaxis, bool calibrateYaxis, bool calibrateZaxis, unsigned int threshold, unsigned int samples)
{
  int16_t sum[3]     = {0,0,0};
  int16_t prevVal[3] = {0,0,0};
  int16_t * raw;  
  int moveCounter 	= 0;
  
  for(int i=0; i<samples;i++)
  {
    raw = _driver.measure();

    /// Detect and count movement
    if (abs(prevVal[0] - raw[0]) > threshold || abs(prevVal[1] - raw[1]) > threshold || abs(prevVal[2] - raw[2]) > threshold)
    {
      moveCounter++;
    }
    else
    {
      sum[0] += raw[0];
      sum[1] += raw[1];
      sum[2] += raw[2];
    }
    prevVal[0] = raw[0];
    prevVal[1] = raw[1];
    prevVal[2] = raw[2];
    delay(1);  /// must wait due to fast loop, sleeping is permitted here since this is an initation routine
    
    /// if 10% movement has occurred, discontinue calibration
    if ((float)moveCounter/samples >= 0.1)
    {
    	return false;
    }
  }
  if (calibrateXaxis) _bias[0] = sum[0] / (float) samples;
  if (calibrateYaxis) _bias[1] = sum[1] / (float) samples;
  if (calibrateZaxis) _bias[2] = sum[2] / (float) samples;
  
  return true;
}