/*
Generalized class for obtaining temperature from Bosch BMP085 sensor.
Written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com)

This program has been written according to the datasheet and uses the same algortihm as suggested.
Because the same algorithm was used, this code appears quite ugly!
*/

#include <Arduino.h>  /// delay()
#include "BMP085.h"



void BMP085::initialize()
{ 
  this->_pressureOffset   = 101325;
  this->_UT               = 0;
  this->_UP               = 0;
  this->_temperature      = 1;
  this->_pressure         = 1;
  
  //Wire.begin();       /// join i2c bus
  this->calibrate();  /// download calibration data
  
  /// load values start values due to the need of delay()
  this->requestTemperature();
  delay(5);            /// delay is okay since begin() is only called once
  this->readTemperature();
  
  this->requestPressure();
  delay(5);
  this->readPressure();
  
  this->calcTemperature();
  this->calcPressure();
  this->requestTemperature();
}



void BMP085::calibrate()
{
  uint8_t * buffer = _i2c.read(_DEVICE, _AC1_MSB, 22);
  
  this->_AC1 = (short) (buffer[0] << 8) | buffer[1];
  this->_AC2 = (short) (buffer[2] << 8) | buffer[3];
  this->_AC3 = (short) (buffer[4] << 8) | buffer[5];
  this->_AC4 = (unsigned short) (buffer[6] << 8) | buffer[7];
  this->_AC5 = (unsigned short) (buffer[8] << 8) | buffer[9];
  this->_AC6 = (unsigned short) (buffer[10] << 8) | buffer[11];
  this->_b1  = (short) (buffer[12] << 8) | buffer[13];
  this->_B2  = (short) (buffer[14] << 8) | buffer[15];
  this->_MB  = (short) (buffer[16] << 8) | buffer[17];
  this->_MC  = (short) (buffer[18] << 8) | buffer[19];
  this->_MD  = (short) (buffer[20] << 8) | buffer[21];
}


/// tell sensor that we want temperature output by writing temperature value to register 
void BMP085::requestTemperature()
{
  _i2c.write(_DEVICE, _REQUEST_REGISTER, _TEMPERATURE_REQUEST);  /// have to wait at least 4.5ms before we can read registers
}
void BMP085::requestPressure()
{
  _i2c.write(_DEVICE, _REQUEST_REGISTER, _PRESSURE_REQUEST);  /// have to wait at least 4.5ms before we can read registers
}


/// remember to wait at least 4.5ms before calling this function after calling requestTemperature()
long BMP085::readTemperature()
{
  uint8_t * buffer = _i2c.read(_DEVICE, _REQUEST_MSB, 2);
  
  this->_UT = (buffer[0] << 8) | buffer[1];  // MSB + LSB

  return this->_UT;
}


/// wait 4.5ms
long BMP085::readPressure()
{
  uint8_t * buffer = _i2c.read(_DEVICE, _REQUEST_MSB, 3);
  
  this->_UP = (((unsigned long) buffer[0] << 16) + ((unsigned long) buffer[1] << 8) + (unsigned long) buffer[2]) >> (8-this->_OSS);

  return this->_UP;
}



void BMP085::calcTemperature()
{
  long x1 = ((this->_UT - (long)this->_AC6)*(long)this->_AC5) >> 15;
  long x2 = ((long)this->_MC << 11)/(x1 + (long)this->_MD);
  this->_B5 = x1 + x2;
  
  this->_temperature = (float) ((this->_B5 + 8) >> 4)/10 + this->_temperatureCalibrate;
}



float BMP085::temperature()
{
  update();
  return this->_temperature;
}



void BMP085::calcPressure()
{
  long B6 = this->_B5 - 4000;
  long x1 = (this->_B2 * ((B6*B6) >> 12)) >> 11;
  long x2 = (this->_AC2 * B6) >> 11;
  long x3 = x1 + x2;
  
  long B3 = ((this->_AC1*4 + x3) << this->_OSS + 2) >> 4;
  x1 = (this->_AC3 * B6) >> 13;
  x2 = (this->_b1 * ((B6*B6) >> 12)) >> 16;
  x3 = (x1 + x2 + 2) / 4;
  
  unsigned long B4 = (this->_AC4 * (unsigned long)(x3 + 32768)) >> 15;
  unsigned long B7 = ((unsigned long)this->_UP - B3) * (50000 >> _OSS);
  
  if (B7 < 0x80000000)
  {
    this->_pressure = (B7*2)/B4;
  }
  else
  {
    this->_pressure = (B7/B4)*2;
  }
  
  x1 = ((this->_pressure >> 8) * (this->_pressure >> 8) * 3038) >> 16;
  x2 = (-7357 * this->_pressure) >> 16;
  
  this->_pressure += (x1 + x2 + 3791) >> 4;
}



float BMP085::pressure()
{
  update();
  return (float) this->_pressure;
}



void BMP085::update()
{
  long milliseconds = millis();
  
  /// make sure the last register write was done for 4.5ms ago
  if (milliseconds - this->_time > 4.5)
  {
    if (!this->_state)  /// true indicate temperature reading
    {
      this->readTemperature();
      this->requestPressure();
    }
    else
    {
      this->readPressure();
      this->calcTemperature();
      this->calcPressure();
      this->requestTemperature();
    }
    this->_time = milliseconds;
    this->_state = !this->_state;
  }
}
