/*
HMC5883L 3-axis digital compass driver core
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) written during winter 2013
*/

#include "HMC5883L.h"
#include <math.h>	/// fmin()
#include <Arduino.h>    /// delay()


void HMC5883L::initialize()
{
  _i2c.write(_DEVICE, _REG_A, 0x70);	/// what the hell does this do!?  
  
  /// have to populate this way to make sure it works in C++ and C++11
  int     a[8] = { 1370, 1090, 820,   660,  440,  390,  330,  230 };
  uint8_t b[8] = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07 };
  for (int i=0;i<8;i++)
  {
    _gainGauss[i] = a[i];
    _gainAddress[i] = b[i];
  }
}


int16_t * HMC5883L::measure()
{
  uint8_t * b = _i2c.read(_DEVICE, _DATA_OUTPUT_X_MSB, 6);  /// read from DATA_OUTPUT_X_MSB and the next 5 registers
  
  this->_raw[0] = (b[0] << 8) | b[1];  /// X
  this->_raw[2] = (b[2] << 8) | b[3];  /// Z -  note the order
  this->_raw[1] = (b[4] << 8) | b[5];  /// Y
  
  /// set orientation
  this->_raw[1] *= -1;
  this->_raw[2] *= -1;
   
  return _raw;
}


void HMC5883L::setGain(int index)
{
  _i2c.write(_DEVICE, _REG_B, _gainGauss[index]);  /// array may run out of index
}


/// put compass in continuous measuring mode
void HMC5883L::continuous()
{
  _i2c.write(_DEVICE, _MODE_REGISTER, _CONTINUOUS);
}


/// put compass in periodic mode, only measure stepwise
void HMC5883L::periodic()
{
  _i2c.write(_DEVICE, _MODE_REGISTER, _PERIODIC);
}


/// put compass in idle mode
void HMC5883L::idle()
{
  _i2c.write(_DEVICE, _MODE_REGISTER, _IDLE);
}


void HMC5883L::removeBias()
{
  _i2c.write(_DEVICE, _REG_A, _BIAS_NONE);
}


void HMC5883L::applyPositiveBias()
{
  _i2c.write(_DEVICE, _REG_A, _BIAS_POSITIVE);  /// 0x10 is default value of CRA and by adding 1 we put 0 and 1 in MS1 and MS0, respectively
}


void HMC5883L::applyNegativeBias()
{
  _i2c.write(_DEVICE, _REG_A, _BIAS_NEGATIVE);  /// by adding 2 we put 1 and 1 in MS1 and MS0, respectively
}



/// Written according to self test procedure given in datasheet
bool HMC5883L::selfTest(int samples, int startGain)
{
  _i2c.write(_DEVICE, _REG_A, 0x71);
  int gain;
  
  int16_t * buffer;

  /// loop through all gains in case of saturation
  for (gain = startGain; gain < 8; gain++)
  {
    long int lowLimit  = _SELF_TEST_POSITIVE_LOW *_gainGauss[gain];
    long int highLimit = _SELF_TEST_POSITIVE_HIGH*_gainGauss[gain];
    long int total[3] = {0,0,0};
    
    setGain(gain);
    continuous();
    delay(6);
    
    applyPositiveBias();
    for (int i=0;i<samples;i++)
    {
      buffer = measure();
      
      total[0] += buffer[0];
      total[1] += buffer[1];
      total[2] += buffer[2];
      delay(1);
    }

    int sum[3];
    sum[0] = total[0]/(samples*1);
    sum[1] = total[1]/(samples*1);
    sum[2] = total[2]/(samples*1);

    if (sum[0] >= lowLimit && sum[0] <= highLimit && sum[1] >= lowLimit && sum[1] <= highLimit && sum[2] >= lowLimit && sum[2] <= highLimit)
    {
      removeBias();	
      return true;  /// saturation not found
    }
    
    /// saturation found, increase gain
  }
  
  return false;
}
