/*
generalized class for obtaining compass measurements
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com)
*/

#include <stdint.h>  /// uint8_t
#include "/home/v/Dropbox/Software/arduino/I2C_/I2C.h"


class HMC5883L
{
  private:
    static const uint8_t  _DEVICE         = 0x1E;    // address to the device
    
    /// The table below shows nominal gain settings. Use the “Gain” column to convert counts to Gauss. Choose a lower gain value (higher GN#) when total field strength causes overflow in one of the data output registers (saturation).
    static const uint8_t  _REG_A          = 0x00;    /// read/write

    
    /// rate
    static const uint8_t  _REG_B          = 0x01;    /// read/write - Configuration Register B for configurating output frequency
    static const uint8_t  _HZ_075         = 0x00;    /// 0.7 5Hz
    static const uint8_t  _HZ_1_5         = 0x01;    /// 1.5 Hz
    static const uint8_t  _HZ_3           = 0x02;    /// 3 Hz
    static const uint8_t  _HZ_7_5         = 0x03;    /// 7.5 Hz
    static const uint8_t  _HZ_15          = 0x04;    /// 15 Hz (default)
    static const uint8_t  _HZ_30          = 0x05;    /// 30 Hz
    static const uint8_t  _HZ_75          = 0x06;    /// 75 Hz

    static const uint8_t  _MODE_REGISTER  = 0x02;    /// read/write
    static const uint8_t  _CONTINUOUS     = 0x00;    /// Continuous-Measurement Mode. In continuous-measurement mode, the device continuously performs measurements and places the result in the data register.
                                                 /// RDY goes high when new data is placed in all three registers. After a power-on or a write to the mode or configuration register,
                                                 /// the first measurement set is available from all three data output registers after a period of 2/fDO and subsequent measurements are available at a frequency of fDO, where fDO is the frequency of data output.
    static const uint8_t  _PERIODIC         = 0x01;    /// Single-Measurement Mode (Default). When single-measurement mode is selected, device performs a single measurement, sets RDY high and returned to idle mode. Mode register returns to idle mode
                                                 /// bit values. The measurement remains in the data output register and RDY remains high until the data output register is read or another measurement is performed.
    static const uint8_t  _IDLE           = 0x02;    /// Idle Mode. Device is placed in idle mode.

    /// the order X, Z, Y is correct in this matter
    static const uint8_t  _DATA_OUTPUT_X_MSB          = 0x03;  /// read
    static const uint8_t  _DATA_OUTPUT_X_LSB          = 0x04;  /// read
    static const uint8_t  _DATA_OUTPUT_Z_MSB          = 0x05;  /// read
    static const uint8_t  _DATA_OUTPUT_Z_LSB          = 0x06;  /// read
    static const uint8_t  _DATA_OUTPUT_Y_MSB          = 0x07;  /// read
    static const uint8_t  _DATA_OUTPUT_Y_LSB          = 0x08;  /// read
    static const uint8_t  _STATUS_REGISTER            = 0x09;  /// read
    static const uint8_t  _IDENTIFICATION_REGISTER_A  = 0x0A;  /// read
    static const uint8_t  _IDENTIFICATION_REGISTER_B  = 0x0B;  /// read
    
    /// bias
    static const uint8_t  _BIAS_NONE      = 0x10;
    static const uint8_t  _BIAS_POSITIVE  = 0x11;
    static const uint8_t  _BIAS_NEGATIVE  = 0x12;
 
    #if defined(__MK20DX128__)
    static constexpr float _SELF_TEST_X_AXIS           = 1.16;
    static constexpr float _SELF_TEST_Y_AXIS           = 1.16;
    static constexpr float _SELF_TEST_Z_AXIS           = 1.08;
    static constexpr float _SELF_TEST_POSITIVE_LOW     = (243/390);  /// 243 is given at gain=5, thus 390.
    static constexpr float _SELF_TEST_POSITIVE_HIGH    = (575/390);
    static constexpr float _SELF_TEST_NEGATIVE_LOW     = -(243/390);
    static constexpr float _SELF_TEST_NEGATIVE_HIGH    = -(575/390);
    #else
    static const float _SELF_TEST_X_AXIS           = 1.16;
    static const float _SELF_TEST_Y_AXIS           = 1.16;
    static const float _SELF_TEST_Z_AXIS           = 1.08;
    static const float _SELF_TEST_POSITIVE_LOW     = (243/390);  /// 243 is given at gain=5, thus 390.
    static const float _SELF_TEST_POSITIVE_HIGH    = (575/390);
    static const float _SELF_TEST_NEGATIVE_LOW     = -(243/390);
    static const float _SELF_TEST_NEGATIVE_HIGH    = -(575/390);
    #endif
    
    
  private:
    I2C     _i2c;
    int16_t _raw[3];
    int16_t _scale[3];
    int     _gainGauss[8];//   = { 1370, 1090, 820,   660,  440,  390,  330,  230 };
    uint8_t _gainAddress[8];// = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07 };


  
  public:
    void initialize();
    int16_t * measure();
    
    void setGain(int index = 1);
    void continuous();
    void periodic();
    void idle();
    
    void removeBias();
    void applyPositiveBias();
    void applyNegativeBias();
    
    bool selfTest(int samples = 64, int startGain = 0);
};
