/*
Madgwick's Attitude and Heading Reference System
implemented in C++ by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during winter 2013
*/

#include "AHRS.h"
#include <math.h>  /// sqrt(), atan2(), asin()


/// Constructor - initialize
AHRS::AHRS(float frequency, float gain)
{
  _q[0] = 1;
  _q[1] = 0;
  _q[2] = 0;
  _q[3] = 0;
  
  _gain  = gain;
  _dt    = 1.0f/frequency;
}


/// Algorithm which incorporates Angular Rate and Gravity (ARG)
void AHRS::ARG(float ax, float ay, float az, float gx, float gy, float gz)
{
	/// Temporary variables
	float f0, f1, f2, f3;
	float _2q0, _2q1, _2q2, _2q3, _4q0, _4q1, _4q2 ,_8q1, _8q2, q0q0, q1q1, q2q2, q3q3;

	computeQdot(gx, gy, gz);

	/// make sure acceleroemter data is valid or NaN will break further calculationsd
	if(ax != 0.0f && ay != 0.0f && az != 0.0f)
	{
		/// Normalize accelerometer measurement
		normalize(ax, ay, az); 

		/// Auxiliary variables to avoid repeated arithmetic
		_2q0 = 2.0f * _q[0];
		_2q1 = 2.0f * _q[1];
		_2q2 = 2.0f * _q[2];
		_2q3 = 2.0f * _q[3];
		_4q0 = 4.0f * _q[0];
		_4q1 = 4.0f * _q[1];
		_4q2 = 4.0f * _q[2];
		_8q1 = 8.0f * _q[1];
		_8q2 = 8.0f * _q[2];
		q0q0 = _q[0] * _q[0];
		q1q1 = _q[1] * _q[1];
		q2q2 = _q[2] * _q[2];
		q3q3 = _q[3] * _q[3];

		/// Gradient decent algorithm corrective step
		f0 = _4q0 * q2q2 + _2q2 * ax + _4q0 * q1q1 - _2q1 * ay;
		f1 = _4q1 * q3q3 - _2q3 * ax + 4.0f * q0q0 * _q[1] - _2q0 * ay - _4q1 + _8q1 * q1q1 + _8q1 * q2q2 + _4q1 * az;
		f2 = 4.0f * q0q0 * _q[2] + _2q0 * ax + _4q2 * q3q3 - _2q3 * ay - _4q2 + _8q2 * q1q1 + _8q2 * q2q2 + _4q2 * az;
		f3 = 4.0f * q1q1 * _q[3] - _2q1 * ax + 4.0f * q2q2 * _q[3] - _2q2 * ay;
		
		/// Normalize gradient
		normalize(f0, f1, f2, f3);

		/// Apply feedback step
		_qDot[0] -= _gain * f0;
		_qDot[1] -= _gain * f1;
		_qDot[2] -= _gain * f2;
		_qDot[3] -= _gain * f3;
	}

	updateQuarternions();	
}


/// Algorithm which incorporates Magnetic, Angular Rate and Gravity
void AHRS::MARG(float ax, float ay, float az, float gx, float gy, float gz, float mx, float my, float mz)
{
	float f0, f1, f2, f3;
	float hx, hy;
	float _2q0mx, _2q0my, _2q0mz, _2q1mx, _2bx, _2bz, _4bx, _4bz, _2q0, _2q1, _2q2, _2q3, _2q0q2, _2q2q3, q0q0, q0q1, q0q2, q0q3, q1q1, q1q2, q1q3, q2q2, q2q3, q3q3;

	/// Use IMU algorithm if magnetometer measurement is invalid or NaN will break further calculations
	if(mx == 0.0f && my == 0.0f && mz == 0.0f)
	{
		ARG(ax, ay, az, gx, gy, gz);
		return;
	}

	computeQdot(gx, gy, gz);

	/// Compute feedback only if accelerometer measurement is valid (avoids NaN)
	if(ax != 0.0f && ay != 0.0f && az != 0.0f)
	{
		/// Normalize accelerometer measurement
		normalize(ax, ay, az);  

		/// Normalize magnetometer measurement
		normalize(mx, my, mz);

		/// Auxiliary variables to avoid repeated arithmetic
		_2q0 = 2.0f * _q[0];
		_2q1 = 2.0f * _q[1];
		_2q2 = 2.0f * _q[2];
		_2q3 = 2.0f * _q[3];
		_2q0mx = _2q0 * mx;
		_2q0my = _2q1 * my;
		_2q0mz = _2q2 * mz;
		_2q1mx = _2q3 * mx;
		_2q0q2 = _2q0 * _q[2];
		_2q2q3 = _2q2 * _q[3];
		q0q0 = _q[0] * _q[0];
		q0q1 = _q[0] * _q[1];
		q0q2 = _q[0] * _q[2];
		q0q3 = _q[0] * _q[3];
		q1q1 = _q[1] * _q[1];
		q1q2 = _q[1] * _q[2];
		q1q3 = _q[1] * _q[3];
		q2q2 = _q[2] * _q[2];
		q2q3 = _q[2] * _q[3];
		q3q3 = _q[3] * _q[3];

		/// Reference direction of Earth's magnetic field
		hx = mx * q0q0 - _2q0my * _q[3] + _2q0mz * _q[2] + mx * q1q1 + _2q1 * my * _q[2] + _2q1 * mz * _q[3] - mx * q2q2 - mx * q3q3;
		hy = _2q0mx * _q[3] + my * q0q0 - _2q0mz * _q[1] + _2q1mx * _q[2] - my * q1q1 + my * q2q2 + _2q2 * mz * _q[3] - my * q3q3;
		_2bx = sqrt(hx * hx + hy * hy);
		_2bz = -_2q0mx * _q[2] + _2q0my * _q[1] + mz * q0q0 + _2q1mx * _q[3] - mz * q1q1 + _2q2 * my * _q[3] - mz * q2q2 + mz * q3q3;
		_4bx = 2.0f * _2bx;
		_4bz = 2.0f * _2bz;

		/// Gradient decent algorithm corrective step
		f0 = -_2q2 * (2.0f * q1q3 - _2q0q2 - ax) + _2q1 * (2.0f * q0q1 + _2q2q3 - ay) - _2bz * _q[2] * (_2bx * (0.5f - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) + (-_2bx * _q[3] + _2bz * _q[1]) * (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) + _2bx * _q[2] * (_2bx * (q0q2 + q1q3) + _2bz * (0.5f - q1q1 - q2q2) - mz);
		f1 = _2q3 * (2.0f * q1q3 - _2q0q2 - ax) + _2q0 * (2.0f * q0q1 + _2q2q3 - ay) - 4.0f * _q[1] * (1 - 2.0f * q1q1 - 2.0f * q2q2 - az) + _2bz * _q[3] * (_2bx * (0.5f - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) + (_2bx * _q[2] + _2bz * _q[0]) * (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) + (_2bx * _q[3] - _4bz * _q[1]) * (_2bx * (q0q2 + q1q3) + _2bz * (0.5f - q1q1 - q2q2) - mz);
		f2 = -_2q0 * (2.0f * q1q3 - _2q0q2 - ax) + _2q3 * (2.0f * q0q1 + _2q2q3 - ay) - 4.0f * _q[2] * (1 - 2.0f * q1q1 - 2.0f * q2q2 - az) + (-_4bx * _q[2] - _2bz * _q[0]) * (_2bx * (0.5f - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) + (_2bx * _q[1] + _2bz * _q[3]) * (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) + (_2bx * _q[0] - _4bz * _q[2]) * (_2bx * (q0q2 + q1q3) + _2bz * (0.5f - q1q1 - q2q2) - mz);
		f3 = _2q1 * (2.0f * q1q3 - _2q0q2 - ax) + _2q2 * (2.0f * q0q1 + _2q2q3 - ay) + (-_4bx * _q[3] + _2bz * _q[1]) * (_2bx * (0.5f - q2q2 - q3q3) + _2bz * (q1q3 - q0q2) - mx) + (-_2bx * _q[0] + _2bz * _q[2]) * (_2bx * (q1q2 - q0q3) + _2bz * (q0q1 + q2q3) - my) + _2bx * _q[1] * (_2bx * (q0q2 + q1q3) + _2bz * (0.5f - q1q1 - q2q2) - mz);
		
		/// Normalize gradient
		normalize(f0, f1, f2, f3);

		/// Apply feedback step
		_qDot[0] -= _gain * f0;
		_qDot[1] -= _gain * f1;
		_qDot[2] -= _gain * f2;
		_qDot[3] -= _gain * f3;
	}

	updateQuarternions();
}


/// Compute q-dot, rate of change,  since last measurement from gyroscope
void AHRS::computeQdot(float gx, float gy, float gz)
{
	/// qDot = 0.5f * q (x) [ 0 gx gy gz ]; where (x) is sign for quaternion product
	_qDot[0] = 0.5f * (-_q[1] * gx - _q[2] * gy - _q[3] * gz);
	_qDot[1] = 0.5f * (_q[0]  * gx + _q[2] * gz - _q[3] * gy);
	_qDot[2] = 0.5f * (_q[0]  * gy - _q[1] * gz + _q[3] * gx);
	_qDot[3] = 0.5f * (_q[0]  * gz + _q[1] * gy - _q[2] * gx);
}


/// Compute new quarternions by integrating q-dot
void AHRS::updateQuarternions()
{
	/// Integrate qDot
	_q[0] += _qDot[0] * _dt;
	_q[1] += _qDot[1] * _dt;
	_q[2] += _qDot[2] * _dt;
	_q[3] += _qDot[3] * _dt;
	
	normalize(_q[0], _q[1], _q[2], _q[3]);
}


/// Fast inverse square-root - http://en.wikipedia.org/wiki/Fast_inverse_square_root
float AHRS::fastInvSqrt(float x)
{
	float halfx = 0.5f * x;
	float y = x;
	long i = *(long*)&y;
	i = 0x5f3759df - (i>>1);
	y = *(float*)&i;
	y = y * (1.5f - (halfx * y * y));
	return y;
}

/// Normalize 3 sized vector
void AHRS::normalize(float &a, float &b, float &c)
{
  float secondNorm = fastInvSqrt(a*a + b*b + c*c);
  
  a *= secondNorm;
  b *= secondNorm;
  c *= secondNorm;
}

/// Normalize 4 sized vector
void AHRS::normalize(float &a, float &b, float &c, float &d)
{
  float secondNorm = fastInvSqrt(a*a + b*b + c*c + d*d);
  
  a *= secondNorm;
  b *= secondNorm;
  c *= secondNorm;
  d *= secondNorm;
}


/// Quaternions
float AHRS::q0()
{
  return _q[0];
}
float AHRS::q1()
{
  return _q[1];
}
float AHRS::q2()
{
  return _q[2];
}
float AHRS::q3()
{
  return _q[3];
}


/// Euler angles
float AHRS::phi()
{
  return atan2(2 * (_q[0]*_q[1] + _q[2]*_q[3]), 1 - 2*(_q[1]*_q[1] + _q[2]*_q[2]) );
}
float AHRS::theta()
{
  return asin(2 * (_q[0]*_q[2] - _q[3]*_q[1]) );
}
float AHRS::psi()
{
  return atan2(2 * (_q[0]*_q[3] + _q[1]*_q[2]), 1 - 2*(_q[2]*_q[2] + _q[3]*_q[3]) );
}
