/*
Generalised class to handle navigation tasks and GPS receiver commnication. Thus communication with NMEA parser (TinyGPS)
Written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during winter 2013
*/

#include "TinyGPS.h"
#include "Navigation.h"


/// Start communication with GPS receiver connected to Serial1
bool Navigation::begin(int baudRate)
{
  Serial1.begin(baudRate);
  return true;
}


/// Add a way-point to way-point stack
void Navigation::addWaypoint(float latitude, float longitude, float altitude)
{
  _waypoint[_numberOfWaypoints] = { latitude, longitude, altitude };
  _numberOfWaypoints++;
}



void Navigation::clearWaypoints()
{
  _numberOfWaypoints = 0;
  _waypointCounter = 0;
}


/// Return true when new position is received
bool Navigation::update()
{
  if (!Serial1.available())
  {
    return false;
  }
  
  int c = Serial1.read();
    
  #ifdef NMEA
  Serial.write(c);
  #endif
    
  /// a NMEA line is ready to be parsed
  if (_GPS.encode(c))
  {
    /// new data available - parse position from NMEA and save to latitude and longitude
    _GPS.f_get_position(&_location.latitude, &_location.longitude, &_age);
    
    return true;
  }
  
  return false;
}


/// Should we iterate to next waypoint?
bool Navigation::next(int threshold)  /// threshold in meters
{
  /// When distance between current position and desired waypoint is less than threshold
  if (_numberOfWaypoints > 0 && distance(_location.latitude, _location.longitude, _waypointTo.latitude, _waypointTo.longitude) <= threshold)
  {
      return true;
  }
  return false;
}


/// Get distance from one geodetic coordinate to another using Haversine formula
float Navigation::distance(float latitude, float longitude, float latitudeTo, float longitudeTo)
{
  float R = 6371000;  // Earth radius in meters
  float dLat = radians(latitudeTo - latitude);
  float dLon = radians(longitudeTo - longitude);
  float lat1 = radians(latitude);
  float lat2 = radians(latitudeTo);

  float a = sin(dLat/2) * sin(dLat/2) + sin(dLon/2) * sin(dLon/2) * cos(lat1) * cos(lat2); 
  float c = 2 * atan2(sqrt(a), sqrt(1-a)); 
  float d = R * c;
  
  return d;
}


/// Return true if next waypoint get loaded
bool Navigation::loadNextWaypoint()
{
  if (_numberOfWaypoints == _waypointCounter)
  {
    /// all waypoints has been loaded, loading next is unsuccessful
    return false;
  }
  
  _waypointTo = _waypoint[_waypointCounter];
  _waypointCounter++;
  
  return true;
}


/// Return geodetic coordinates of last known position
float Navigation::latitude()
{
  return _location.latitude;
}
float Navigation::longitude()
{
  return _location.longitude;
}
float Navigation::altitude()
{
  return _GPS.f_altitude();return _location.altitude;
}


/// Return geodetic coordinates of current desired way-point
float Navigation::latitudeTo()
{
  return _waypointTo.latitude;
}
float Navigation::longitudeTo()
{
  return _waypointTo.longitude;
}
float Navigation::altitudeTo()
{
  return _waypointTo.altitude;
}


/// Set origin/home coordinates
void Navigation::setOrigin(float latitude, float longitude, float altitude)
{
  _origin = { latitude, longitude, altitude };
}


/// Navigate home
void Navigation::home()
{
  /// Decrease waypoint counter in case home flying is aborted
  if (_waypointCounter > 0)
  {
    _waypointCounter--;
  }
  
  _waypointTo = _origin;
}


/// Return if navigating home
bool Navigation::toHome()
{
  if (_waypointTo.latitude != _origin.latitude)
  {
    return false;
  }
  if (_waypointTo.longitude != _origin.longitude)
  {
    return false;
  }
  if (_waypointTo.altitude != _origin.altitude)
  {
    return false;
  }
  return true;
}


/// Return current bearing for course
float Navigation::course()
{
  float lat1 = _location.latitude;
  float lat2 = _waypointTo.latitude;
  float long1 = _location.longitude;
  float long2 = _waypointTo.longitude;
  // returns course in degrees (North=0, East=90, South=180, West=270) from position 1 to position 2,
  // both specified as signed decimal-degrees latitude and longitude.
  // Because Earth is no exact sphere, calculated course may be off by a tiny fraction.
  // Courtesy of Maarten Lamers
  float dlon = radians(long2-long1);
  lat1 = radians(lat1);
  lat2 = radians(lat2);
  float a1 = sin(dlon) * cos(lat2);
  float a2 = sin(lat1) * cos(lat2) * cos(dlon);
  a2 = cos(lat1) * sin(lat2) - a2;
  a2 = atan2(a1, a2);
  if (a2 < 0.0)
  {
    a2 += TWO_PI;
  }
  return degrees(a2);
}


/// Convert geodetic (latitude,longitude) to Cartesian coordinates
void Navigation::geodetic2cartesian(long latitude, long longitude, long height)
{
  long N = _r_e^2 / (long)sqrt( _r_e^2 * (long)(cos(latitude))^2 + _r_p^2 * (long)(sin(latitude))^2 );
  
  long x_e = (N + height) * cos(latitude) * cos(longitude);
  long y_e = (N + height) * cos(latitude) * sin(longitude);
  long z_e = ( (_r_p^2/_r_e^2) * N + height) * sin(latitude);
}


/// Number of connected satellites
int Navigation::satellites()
{
  return _GPS.satellites() == TinyGPS::GPS_INVALID_SATELLITES ? 0 : _GPS.satellites();
}

