#include <Arduino.h>
#include "Receiver.h"


/// allocate static variables
PWM Receiver::_pwm[6];
bool Receiver::_lostLink;
bool Receiver::_attached;
int Receiver::_throttlePin;
int Receiver::_aileronPin;
int Receiver::_elevatorPin;
int Receiver::_modePin;


/// Tell which pin we are to assign for interrupt
void Receiver::attach(int throttlePin, int aileronPin, int elevatorPin, int modePin)
{
  _throttlePin = throttlePin;
  _aileronPin  = aileronPin;
  _elevatorPin = elevatorPin;
  _modePin     = modePin;
  _attached 	= true;
  _lostLink    = false;

  _pwm[throttlePin].add(throttlePin, &Receiver::throttleRising,   900, 500);
  _pwm[aileronPin].add(aileronPin,   &Receiver::aileronRising,   1500, 500);
  _pwm[elevatorPin].add(elevatorPin, &Receiver::elevatorRising,  1500, 500); 

  /// The mode pin is not required to be attached
  if (modePin > 0)
  {
    _pwm[modePin].add(modePin, &Receiver::modeRising, 1000, 500);
  }
}


void Receiver::detach(int throttlePin, int aileronPin, int elevatorPin, int modePin)
{
  _attached = false;
  _pwm[throttlePin].remove(throttlePin);
  _pwm[aileronPin].remove(aileronPin);
  _pwm[elevatorPin].remove(elevatorPin);
  
  if (modePin > 0)
  {
     _pwm[modePin].remove(modePin);
  }
}

bool Receiver::attached()
{
  return _attached;
}

int Receiver::signal(int pin)
{
  return _pwm[pin].signal();
}

int Receiver::angle(int pin)
{
  return _pwm[pin].angle();
}


/// Use this function to check whether the receiver has lost with transmitter
bool Receiver::lostLink(int pin)
{
  /// This function may be very custom depending on RC-system. On Spektrum systems, the same throttle position you used when binding to the receiver will be applied when signal is lost.
  /// It is therefore wise to push throttle signal below normal operating range when binding. By doing this, a throttle signal was measured to be about 924 microseconds, hence 1000 value.
  
  if (_pwm[pin].signal() < 1000 || _pwm[pin].signal() > _pwm[pin].maxPulseWidth())
  {
    _lostLink = true;  /// use recoverLink() to change state
    return true;
  }
  
  return false;
}


/// Tell whether signal has been lost and the link has become active
bool Receiver::recoverLink(int pin)
{
  /// if lost link before and not now
  if (_lostLink && !lostLink(pin))
  {
    _lostLink = false;
    return true;
  }
  
  return false;
}


/// Function called when a pin change state from low to high
void Receiver::throttleRising()
{
  _pwm[_throttlePin].handleRising(_throttlePin, Receiver::throttleFalling);
}
/// Function is called when a pin state changes from high to low
void Receiver::throttleFalling()
{
  _pwm[_throttlePin].handleFalling(_throttlePin, Receiver::throttleRising);
}
void Receiver::aileronRising()
{
  _pwm[_aileronPin].handleRising(_aileronPin, Receiver::aileronFalling);
}
void Receiver::aileronFalling()
{
  _pwm[_aileronPin].handleFalling(_aileronPin, Receiver::aileronRising);
}
void Receiver::elevatorRising()
{
  _pwm[_elevatorPin].handleRising(_elevatorPin, Receiver::elevatorFalling);
}
void Receiver::elevatorFalling()
{
  _pwm[_elevatorPin].handleFalling(_elevatorPin, Receiver::elevatorRising);
}
void Receiver::modeRising()
{
  _pwm[_modePin].handleRising(_modePin, Receiver::modeFalling);
}
void Receiver::modeFalling()
{
  _pwm[_modePin].handleFalling(_modePin, Receiver::modeRising);
}
