/*
Class to handle properties of a flying wing
Written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during spring 2013


Future improvements:
- For better customization it is adviced that the servo-related stuff is done in an controlled enviroment such that the timers can be modified as wanted. Thus, undependent of Servo.h.
- Using microseconds to control the servoes would give a better resolution than the angle approach used here.
*/



#include "Actuator.h"
#include <stdlib.h>
#include <math.h>

#ifdef __MK20DX128__
  #include "/home/v/teensy/libraries/Servo/Servo.cpp"
#else
  #include "/usr/share/arduino/libraries/Servo/Servo.cpp"
#endif


/// Connect to left balance rudder actuator
void Actuator::attachLeft(int pin)
{
  left.attach(pin);
  this->angleLeft = 90;  /// 90 is neutral - trim can be added here
  left.write(this->angleLeft);
}


/// Connect to right balance rudder actautor
void Actuator::attachRight(int pin)
{
  right.attach(pin);
  this->angleRight = 90;  /// 90 is neutral - trim can be added here
  right.write(this->angleRight);
}


/// Connect to motor ESC
void Actuator::attachMotor(int pin)
{
  motor.attach(pin);
  this->angleMotor = 0;
  motor.write(this->angleMotor);
}


/// Send high pulse length to left rudder actautor
void Actuator::writeMicrosecondsLeft(int microseconds)
{
  left.writeMicroseconds(microseconds);
}


/// Send high pulse length to right rudder actautor
void Actuator::writeMicrosecondsRight(int microseconds)
{
  right.writeMicroseconds(microseconds);
}


/// Send high pulse length to motor ESC
void Actuator::writeMicrosecondsMotor(int microseconds)
{
  motor.writeMicroseconds(microseconds);
}


/// Set desired vehicle orientations
void Actuator::orientations(int phi, int theta, int psi)
{
  this->_phi   = phi;
  this->_theta = theta;
  this->_psi   = psi;
  update();
}


/// Inflict desired actautor angles
void Actuator::update()
{
  /// phi and theta is (should be) between -90 and 90 deg
  int angleLeft  = 90 + this->_phi + this->_psi - this->_theta;  /// 90 is in neutral position
  int angleRight = 90 + this->_phi + this->_psi + this->_theta;  /// 90 is in neutral position
  
  if (angleLeft  < 0)   angleLeft  = 0;
  if (angleRight < 0)   angleRight = 0;
  if (angleLeft  > 180) angleLeft  = 180;
  if (angleRight > 180) angleRight = 180;
  
  /// Send desired movements to acutaors
  left.write(angleLeft);
  right.write(angleRight);
  //motor.write(angleMotor);  /// 0 is neutral position (off)
}


/// Negative angle rolls left and positive rolls right
void Actuator::roll(int angle)
{
  this->_phi = angle;
  update();
}


/// Negative angle pitches up
void Actuator::pitch(int angle)
{
  this->_theta = angle;
  update();
}


/// Changes yaw direction. Note that this angle is relative and not absolute. Therefore, it must be set zero when desired course is reached. 
void Actuator::course(int angle)
{
  this->_psi = angle;
  update();
}


/// Aeroplanes must roll to change course. By specifying a corse angle, this function will make your plane roll in the shortest direction.
void Actuator::rollByCourse(float degree)
{
  if (degree < 0.0)
  {
    degree += 360;
  }
  
  if (degree <= 180.0)
  {
    roll((int)degree/2);   /// roll right
  }
  else
  {
    roll((int) -(degree-180)/2);  /// roll left
  }
}
