/*

Complete fused application for a way-point autonomous flying drone. Change Actuator class to suit your flight frame.
Written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during spring 2013
- adopted by ...

*/


#include "CONFIG.h"

#include "CD/3.AHRS/code/system/Wire.h"
#include "CD/3.AHRS/code/system/I2C.cpp"


///
/// ACCELERATION
///
#include "CD/3.AHRS/code/drivers/accelerometer/Acceleration.cpp"      /// Generalised class for obtaining accelerations from accelerometer
#include "CD/3.AHRS/code/drivers/accelerometer/ADXL345_/ADXL345.cpp"  /// Select driver
Acceleration<ADXL345> acceleration;                                   /// Create object


///
/// ACTUATORS
///
#include "Actuator.h" /// Generalised class for movement handling
Actuator actuate;     /// Create object


///
/// ALTITUDE
///
#include "CD/3.AHRS/code/drivers/altimeter/Altimeter.cpp"       /// Generalised class for obtaining height from pressure and temperature sensors
#include "CD/3.AHRS/code/drivers/altimeter/BMP085_/BMP085.cpp"  /// Select driver
Altimeter<BMP085> altitude;                                     /// Create object with selected driver


///
/// ATTITUDE HEADING REFERENCE SYSTEM
///
#include "CD/3.AHRS/code/system/AHRS.cpp"  /// Attitude and Heading Reference System (filter)
AHRS ahrs( _HERTZ, _GAIN);                 /// Create object with given frequency and gain, respectively


///
/// BUFFER
///
#include "CD/5.Communication/code/Buffer_/Buffer.cpp"  /// Class to buffer from e.g. serial communication before parsing
Buffer buffer;                                         /// Create object


///
/// COMPASS
///
#include "CD/3.AHRS/code/drivers/magnetometer/Compass.cpp"             /// Generalised class for obtaining magnet searching heading or attitude
#include "CD/3.AHRS/code/drivers/magnetometer/HMC5883L_/HMC5883L.cpp"  /// Select driver
Compass<HMC5883L> compass;                                             /// Create object with selected driver


///
/// GYRO
///
#include "CD/3.AHRS/code/drivers/gyro/Gyro.cpp"               /// Generalised class for obtaining inertial movement from gyroscope
#include "CD/3.AHRS/code/drivers/gyro/L3G4200D_/L3G4200D.cpp" /// Select driver
Gyro<L3G4200D> gyro;                                          /// Create object with selected driver


///
/// RECEIVER (INTERRUPT)
///
#include "CD/5.Communication/code/ControlLink/PWM.cpp"        /// Generalised class for parsing pulse width modulated signals
#include "CD/5.Communication/code/ControlLink/Receiver.cpp"   /// Handle signals outgoing from receiver
Receiver receiver;                                            /// Create object


///
/// NAVIGATION
///
#include "CD/4.Navigation/code/TinyGPS.cpp"    /// GPL class for parsing NMEA written by ...not me...
#include "CD/4.Navigation/code/Navigation.cpp" /// Handle navigation
Navigation navigation;         				     /// Create object


///
/// STABILIZATION
///
#include "PID.cpp"
PID roll;
PID pitch;
PID yaw;
PID surge;


/// Make LED blink
void blink(int requestedBlinks = 1, int pause = 100)
{
  pinMode(13,OUTPUT);
  digitalWrite(13,LOW);
  int state = HIGH;
  delay(pause);
  for (int i=0;i<requestedBlinks;i++)
  {
    digitalWrite(13,state);
    state = (state==LOW?HIGH:LOW);
    delay(pause);
  }
  
  digitalWrite(13,LOW);
}


/// Return true if aided mode is enabled
bool aidedMode()
{
  if (receiver.signal(_PIN_MODE) > _SIGNAL_MANUAL_MODE && receiver.signal(_PIN_MODE) < _SIGNAL_WP_MODE)
  {
    return true;
  }
  
  return false;
}


/// Return true if way-point mode is enabled
bool waypointMode()
{
  if (receiver.signal(_PIN_MODE) > _SIGNAL_WP_MODE)
  {
    return true;
  }
  
  return false;
}


/// Compare two strings
bool equal(char * str1, char * str2)
{
  if (strncmp(str1, str2, strlen(str2)) == 0)
  {
    return true;
  }
  
  return false;
}


///
/// HANDLE INCOMMING DATA ON WIRLESS SERIAL
///
void wireless()
{
  /// There are no data available from the transceiver
  if (!Serial2.available())
  {
    return;
  }

  /// Returns true whenever a new line is read
  if (!buffer.until(Serial2.read(),'\n'))
  {
    return;
  }

  char * word1 = strtok(buffer.get()," ");
  
  /// Clear waypoints
  if (equal(word1,"WPC"))
  {
    navigation.clearWaypoints();
    Serial.println("NAVIGATION: removing all way-points");
  }
  
  /// Verify identity
  else if (equal(word1, _IDENTITY))
  {
    char * word2 = strtok(NULL," ");
      
    /// New waypoint
    if (equal(word2,"WPA"))
    {
      navigation.addWaypoint(atof(strtok(NULL, " ")), atof(strtok(NULL, " ")), atof(strtok(NULL, " ")));
      Serial.println("NAVIGATION: new waypoint added");
      
      /// If this is the first point then load it
      if (navigation.numberOfWaypoints() == 1)
      {
        navigation.loadNextWaypoint();
        Serial.println("NAVIGATION: loaded first way-point");
      }
    }
  }
}


///
/// INITIATION
///
void setup()
{
  ///
  /// CONNECT GPS
  ///
  Serial.begin(_SERIAL_BAUD);
  #ifdef __MK20DX128__
  while (!Serial)                                   /// Some (!Atmel) chips require this hack to make serial work with the host
  {
    blink();
  }
  #endif
  Serial.println("USB: serial started");

  
  ///
  /// CONNECT BASE STATION (WIRELESS TRANSCEIVER)
  ///
  Serial2.begin(_WIRELESS_BAUD);                    /// Serial2 means pin 9 and 10 on Teensy 3.0
  Serial.println("WIRELESS: serial started");


  ///
  /// CONNECT GPS RECEIVER
  ///
  navigation.begin(_GPS_BAUD);              /// Navigation handles GPS serial communication on Serial1 for pin 19 (rx) and 18 (tx) for Teensy 3.0
  Serial.println("NAVIGATION: GPS serial started");
  
 
  ///
  /// CALIBRATE AND IMU INITIAL
  ///
  Wire.begin();
  Serial.println("I2C: joined bus");
  
  Serial.println("About to calibrate, do not move!");
  delay(500);


  /// Initialize accelerometer
  acceleration.setHz( _HERTZ );
  acceleration.setG( _MAX_G );
  acceleration.initialize(9.81/269, 9.81/250, 9.81/260, 9.81/260, 9.81/253, 9.81/264);
  while (!acceleration.calibrate(true,true,false,_ACCELEROMETER_NOISE))
  {
    Serial.println("ACCELEROMETER: re-calibrating in 1 second.");
    delay(1000);
  }
  Serial.println("ACCELEROMETER: calibrated");


  /// Initialize gyroscope. Gyro is default in sleep mode and must be set active, this fails sometimes
  while (!gyro.initialize((float) _HERTZ / 4000 ))        /// 800 Hz / 2000 Degree/second * 0.6 (tunable factor)
  {
    Serial.println("GYRO: init failed. Correct device address? Reattempt in 1 second.");
    delay(1000);
  }
  while (!gyro.calibrate(true,true,true,_GYRO_NOISE))
  {
    Serial.println("GYRO: re-calibrating in 1 second.");
    delay(1000);
  }
  Serial.println("GYRO: calibrated");
  

  /// Initialize compass
  compass.initialize();
  /*while (!compass.calibrate(64, 1))
  {
    Serial.println("COMPASS: calibrate failed. Reattempt in 1 second.");
    delay(1000);
  }*/
  compass.setGain(1);
  compass.continuous();
  Serial.println("COMPASS: ready");
  
  delay(500);
  
  ///
  /// DEFINE PID VALUES
  ///
  surge.set(_SURGE_P, _SURGE_I, _SURGE_D);
  roll.set(_PITCH_P, _PITCH_I, _PITCH_D);
  pitch.set(_ROLL_P, _ROLL_I, _ROLL_D);
  yaw.set(_YAW_P, _YAW_I, _YAW_D);
  
  
  ///
  /// CONNECT RECEIVER
  ///
  receiver.attach(_PIN_THROTTLE, _PIN_AILERON, _PIN_ELEVATOR, _PIN_MODE);

 
  ///
  /// CONNECT ACTUATORS
  ///
  #ifndef _DISABLE_ACTUATOR
  actuate.attachLeft( _PIN_LEFT );
  actuate.attachRight( _PIN_RIGHT );
  actuate.attachMotor( _PIN_MOTOR );
  #endif
  
  
  ///
  /// SET ORIGIN
  ///
  /// wait until at least four satellites are connected then another five seconds such that GPS can iterate to a better position or even track more satellites
  /*unsigned long time = millis();
  while ((millis()-time) <= 5000)  /// preferably we would have more than 4, but in cases where only 4 is available the vehicle would be useless
  {
    navigation.update();   
    if (navigation.satellites() < 4)
    {
      time = millis();
    }
    blink(2,10);
    delay(100);
  }

  navigation.setOrigin(navigation.latitude(), navigation.longitude(), navigation.altitude());*/
  
  _LAUNCHED = true;
}


///
/// MAIN LOOP CONTROLLER HANDLERS
///
float phi_d, phi_e;      /// Desired roll angle and difference between them.
float theta_d, theta_e;  /// Desired pitch angle and difference between them.
float psi_d, psi_e;      /// Desired heading angle and difference between them.


///
/// MAIN LOOP TIMERS
///
unsigned long time;
unsigned long pHz01 = 0;
unsigned long pHz1 = 0;
unsigned long pHz10 = 0;
unsigned long pHz100 = 0;
unsigned long pHz800 = 0;


///
/// CONTINUOUS LOOP (MAIN LOOP)
///
void loop()
{
  time = micros();      /// Store time in memory for this iteration. This will overflow in 70 minutes.
  
  /// Overflow compensation
  if (pHz01 > time)
  {
    pHz01 = pHz1 = pHz10 = pHz100 = pHz800 = time;
  }
  
  /// laying-8 Hz
  continuous();
  
  
  /// 0.1 Hz
  if ( (time - pHz01) >= 10000000)
  {
    hz01();
    pHz01 = time;
  }


  /// 1 Hz
  if ( (time - pHz1) >= 1000000)
  {
    hz1();
    pHz1 = time;
  }
 
 
  /// 10 Hz
  if ( (time - pHz10) >= 100000)
  {
    hz10();
    pHz10 = time;
  }
  
 
  /// 100 Hz
  if ( (time - pHz100) >= 10000)
  {   
    hz100();    
    pHz100 = time;
  }
  
  
  /// 800 Hz
  if ( (time - pHz800) >= 1250)
  {
    hz800();
    pHz800 = time;
  }
}


///
/// CONTINUOUS LOOP
///
void continuous()
{
  navigation.update();  /// Update current position using GPS
  wireless();           /// Updaate data link
}


///
/// 0.1 HZ LOOP
///
void hz01()
{
  /// Send vehicle position to base station
  Serial2.print(_IDENTITY);
  Serial2.print(" P ");
  Serial2.print(navigation.latitude(), 8);
  Serial2.print(" ");
  Serial2.print(navigation.longitude(), 8);
  Serial2.print(" ");
  Serial2.println(navigation.altitude(), 8);
}


///
/// 1 HZ LOOP
///
void hz1()
{
  /// Not launched, stop executing
  if (!_LAUNCHED)
  {
    return;
  }
  
  /// If signal is lost, return home
  if (pHz1 != 0 && receiver.lostLink( _PIN_THROTTLE ))
  {  
    Serial.println("RECEIVER: no signal"); 
    
    /// Has been ordered home, ignore future waypoints
    if (!navigation.toHome())
    {
      navigation.home();
      Serial.println("NAVIGATION: flying home");
    }
  }
  /// If signal has been resumed, recover flight path
  else if (receiver.recoverLink( _PIN_THROTTLE ))
  {
    Serial.println("RECEIVER: signal recovered"); 
    navigation.loadNextWaypoint();
  }
}


///
/// 10 HZ LOOP
///
void hz10()
{
  /// Not launched, stop executing
  if (!_LAUNCHED)
  {
    return;
  }

  /// Has been ordered home, ignore future waypoints
  if (navigation.toHome())
  {
    return;
  }

  /// WP-mode is set, check if next waypoint is ready
  if (waypointMode() && navigation.next(_NAVIGATION_THRESHOLD) )
  {
    Serial.println("NAVIGATION: Waypoint reached");
    
    if (navigation.loadNextWaypoint())
    {
      Serial.println("NAVIGATION: New waypoint loaded");
    }
    else
    {
      Serial.println("NAVIGATION: finished all waypoints, flying home");
      navigation.home();
    }
  }
}


///
/// 100 HZ LOOP
///
void hz100()
{    
  /// If either aided control OR wp-mode is selected OR the transmitter is out of range
  if (aidedMode() || waypointMode() || receiver.lostLink( _PIN_THROTTLE ))
  {
    ///
    /// WAYPOINT TRACKING
    ///
    if (waypointMode() || receiver.lostLink( _PIN_THROTTLE ))
    {
      /// Errors
      phi_e   = roll.stabilize( degrees(ahrs.phi())*-1, 0);    /// phi_e = phi - phi_d
      theta_e = pitch.stabilize(degrees(ahrs.theta())*-1, 0);  /// theta_e = theta - theta_d
      
      /// Must have enough satelites to determine own position and course
      if (navigation.satellites() >= 4)
      {
        psi_e = yaw.stabilize(compass.heading(), navigation.course()); /// psi_e = psi - psi_d
      }
      else
      {
        theta_e -= 10;  /// Make sure plane has enough pitch
        psi_e = 20;     /// Fly in circle (left) until enough satellites are available
      }
      
      /// Manual throttle control
      actuate.writeMicrosecondsMotor( receiver.signal( _PIN_THROTTLE ) );
    }
    ///
    /// AIDED CONTROL (CLOSED LOOP)
    ///
    else
    {
      /// errors
      phi_e   = roll.stabilize(degrees(ahrs.phi())*-1, receiver.angle( _PIN_AILERON ) - 90);
      theta_e = pitch.stabilize(degrees(ahrs.theta())*-1, receiver.angle( _PIN_ELEVATOR ) - 90);      
      psi_e = 0;
    }
    
    /// Limit actuating signals
    phi_e   = constrain(phi_e, - _MAX_ROLL_ANGLE, _MAX_ROLL_ANGLE);
    theta_e = constrain(theta_e, _MAX_PITCH_ANGLE*-1, _MIN_PITCH_ANGLE*-1);  /// theta is flipped
    psi_e   = constrain(psi_e, - _MAX_YAW_ANGLE, _MAX_YAW_ANGLE);
    
    /// Actuate rudders according to model based on orientation errors
    actuate.orientations(phi_e, theta_e, psi_e);
  }
  ///
  /// MANUAL CONTROL OF FLYING WING (OPEN LOOP)
  ///
  else
  {
    actuate.writeMicrosecondsMotor( receiver.signal( _PIN_THROTTLE ) );
    actuate.writeMicrosecondsLeft( receiver.signal( _PIN_AILERON ) + (1500-receiver.signal( _PIN_ELEVATOR )) );
    actuate.writeMicrosecondsRight( receiver.signal( _PIN_AILERON ) - (1500-receiver.signal( _PIN_ELEVATOR )) );
  }
  
  #ifdef _ERRORS
  Serial.print(phi_e);
  Serial.print("\t");
  Serial.print(theta_e);
  Serial.print("\t");
  Serial.print(psi_e);
  Serial.println();
  #endif
  
  #ifdef _SIGNALS
  Serial.print(receiver.signal( _PIN_THROTTLE ));
  Serial.print("\t");
  Serial.print(receiver.signal( _PIN_AILERON ) + (1500-receiver.signal( _PIN_ELEVATOR )) );
  Serial.print("\t");
  Serial.print(receiver.signal( _PIN_AILERON ) - (1500-receiver.signal( _PIN_ELEVATOR )) );
  Serial.print("\t");
  Serial.println(receiver.signal( _PIN_MODE ));
  #endif
}



///
/// 800 HZ LOOP
///
void hz800()
{
  /// Download sensors data
  acceleration.measure();
  gyro.measure();
  compass.measure();

  /// Altering gyro measurements due to noise! .. using low pass filter
  if ( abs(gyro.x(0) - gyro.x(1)) > 800)
  {
    gyro.setAxisValue(0,0,gyro.x(1));
  }
  if ( abs(gyro.y(0) - gyro.y(1)) > 800)
  {
    gyro.setAxisValue(1,0,gyro.y(1));
  }
  if ( abs(gyro.z(0) - gyro.z(1)) > 800)
  {
    gyro.setAxisValue(2,0,gyro.z(1));
  }

  
  /// Filter data
  //ahrs.MARG(acceleration.x(), acceleration.y(), acceleration.z(), radians(gyro.x()), radians(gyro.y()), radians(gyro.z()), compass.x(), compass.y(), compass.z());
  ahrs.ARG(acceleration.x(), acceleration.y(), acceleration.z(), radians(gyro.x()), radians(gyro.y()), radians(gyro.z()));


  /// Display Euler angles for visualisation
  #ifdef _SIMULATOR
  Serial.print(ahrs.phi());
  Serial.print("\t");
  Serial.print(ahrs.theta());
  Serial.print("\t");
  Serial.print(ahrs.psi());
  Serial.println();
  #endif
  
  /// Display IMU data
  #ifdef _IMU
  Serial.print(acceleration.x());
  Serial.print("\t");
  Serial.print(acceleration.y());
  Serial.print("\t");
  Serial.print(acceleration.z());
  Serial.print("\t");
  Serial.print(gyro.x());
  Serial.print("\t");
  Serial.print(gyro.y());
  Serial.print("\t");
  Serial.print(gyro.z());
  Serial.println();
  #endif
}

