/*
generalized class for obtaining accelerations from ADXL345
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com)
*/

#include "Acceleration.h"
#include <Wire.h>


void Acceleration::begin()
{
  maxOutput = 510;  /// this accelerometer can only measure 16g = 16*9.81 = 156.96 at most. Output value of 255 represent 9.81m/s^2. Thus, 510 corresponds to 2g.
  
  /// Calibration found once(!) by positioning the accelerometer pefectly level and record readings for this meter it's found to be:
  bias[0] = -0.04;  /// X
  bias[1] = 0.4;    /// Y
  bias[2] = 0.0;    /// Z
  
  /// Set calibration data. These numbers has been found by collecting raw data during slowly varying movements. 9.81 m/s^2 is earths gravetational acceleration.
  calibrationGainMin[0] = -9.81/-269;  /// Smallest measured raw value was -269
  calibrationGainMax[0] = 9.81/250;  /// Highest value 250
  calibrationGainMin[1] = -9.81/-260;
  calibrationGainMax[1] = 9.81/260;  
  calibrationGainMin[2] = -9.81/-253;
  calibrationGainMax[2] = 9.81/264;
  
  Wire.begin();                         /// Join i2c bus
  write(DEVICE, BW_RATE,BW_400);        /// Put sensor in 400Hz mode
  write(DEVICE, DATA_FORMAT,B1000);     /// Put sensor in full resolution and +/- 2g
  write(DEVICE, POWER_CTL, B00001011);  /// Put sensor in measure mode (turn on)
}

void Acceleration::write(uint8_t device, uint8_t address, uint8_t code)
{
  Wire.beginTransmission(device);
  Wire.write(address);
  Wire.write(code);
  Wire.endTransmission();
}


uint8_t * Acceleration::read(uint8_t device, uint8_t address,int numberOfregisters)
{
  /// tell accelerometer which address we would like to access
  Wire.beginTransmission(device);
  Wire.write(address);
  Wire.endTransmission();

  /// get the contents of register
  Wire.beginTransmission(device);
  Wire.requestFrom((int) device, numberOfregisters);

  int i = 0;
  while(Wire.available())
  { 
    this->buffer[i] = Wire.read();  /// receive one uint8_t
    i++;
  }
  Wire.endTransmission();
  
  return buffer;
}

/// Get accelerations from accelerometer (d'oh)
void Acceleration::getRaw()
{ 
  /// there are 6 registers for acceleration data, we start by reading from register DATAX0 and pluss 5
  uint8_t * b = read(DEVICE, DATAX0, 6);
  
  raw[0] = (b[1] << 8) | b[0];
  raw[1] = (b[3] << 8) | b[2];
  raw[2] = (b[5] << 8) | b[4];
}

void Acceleration::get()
{
  getRaw();
  
  /// Raw data gets altered, not pretty!
  raw[0] -= bias[0];
  raw[1] -= bias[1];
  raw[2] -= bias[2];
  
  /// Do not accept unrealistic values
  if (abs(raw[0]) > maxOutput || abs(raw[1]) > maxOutput || abs(raw[2]) > maxOutput)
  {
    return;
  }
  
  /// FIFO stack for INS purposes
  xAxis[2] = xAxis[1];
  xAxis[1] = xAxis[0];
  xAxis[0] = raw[0];
  yAxis[2] = yAxis[1];
  yAxis[1] = yAxis[0];
  yAxis[0] = raw[1];
  zAxis[2] = zAxis[1];
  zAxis[1] = zAxis[0];
  zAxis[0] = raw[2];
  
  /// Convert to m/s^2. Max and min values for each axis is different, therefore different gain will be applied.
  xAxis[0] *= (xAxis[0] < 0 ? calibrationGainMin[0] : calibrationGainMax[0] );
  yAxis[0] *= (yAxis[0] < 0 ? calibrationGainMin[1] : calibrationGainMax[1] );
  zAxis[0] *= (zAxis[0] < 0 ? calibrationGainMin[2] : calibrationGainMax[2] );
}


/// Get acceleration in m/s^2 in x: axis=0, y: axis=1 or z: axis=2 directions
double Acceleration::axis(int axis, int index)
{ 
  if (index > 2)
  {
    index = 0;
  }

  /// Values below are negative because of mounting direction of accelerometer
  if (axis == 1)
  {
    return -yAxis[index];
  }
  else if (axis == 2)
  {
    return -zAxis[index];
  }
  else
  {
    return -xAxis[index];
  }
}


double Acceleration::x(int index)
{
  return axis(0,index);
}
double Acceleration::y(int index)
{
  return axis(1,index);
}
double Acceleration::z(int index)
{
  return axis(2,index);
}


double Acceleration::roll()  /// phi
{
  double angle = atan2(y(),z()) - M_PI;
  if (angle < 0.0)
  {
    angle += TWO_PI;
  }
  return angle;
}


double Acceleration::pitch()  /// theta
{
  double angle = TWO_PI - atan2(x(),z()) - M_PI;
  if (angle < 0.0)
  {
    angle += TWO_PI;
  }
  return angle;
}

/// A custom function since OFSX registers did not work as expected
bool Acceleration::calibrate(bool calibrateXaxis, bool calibrateYaxis, bool calibrateZaxis)
{
  int wrongCounter = 0;
  int16_t sum[3] = {0,0,0};
  for(int i=0; i<128; i++)
  {
    getRaw();
    /// only accept realistic values!
    if (abs(this->raw[0]) > maxOutput || abs(this->raw[1]) > maxOutput || abs(this->raw[2]) > maxOutput)
    {
      i--;
    }
    else
    {
      sum[0] += this->raw[0];
      sum[1] += this->raw[1];
      sum[2] += this->raw[2];
    }
    delay(2);  /// must wait due to fast processor clock
  }
  if (calibrateXaxis)
  {
    bias[0] = sum[0] / 128.0;
  }
  if (calibrateYaxis)
  {
    bias[1] = sum[1] / 128.0;
  }
  if (calibrateZaxis)
  {
    bias[2] = sum[2] / 128.0;
  }
  
  return true;
}

