/*
Generalized class for obtaining temperature from Bosch BMP085 sensor.
Written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com)

This program has been written according to the datasheet and uses the same algortihm as suggested.
*/

#include <Wire.h>
#include "Altitude.h"

Altitude::Altitude()
{
  this->pressureOffset   = 101325;
  this->UT               = 0;
  this->UP               = 0;
  this->temperature      = 1;
  this->pressure         = 1;
}

void Altitude::begin()
{ 
  //Wire.begin();       /// join i2c bus
  this->calibrate();  /// download calibration data
  
  /// load values start values due to the need of delay()
  this->requestTemperature();
  delay(5);            /// delay is okay since begin() is only called once
  this->readTemperature();
  
  this->requestPressure();
  delay(5);
  this->readPressure();
  
  this->trueTemperature();
  this->truePressure();
  this->calculate();
  this->requestTemperature();
}

void Altitude::write(uint8_t address, uint8_t code)
{
  Wire.beginTransmission(DEVICE);
  Wire.write(address);
  Wire.write(code);
  Wire.endTransmission();
}


uint8_t * Altitude::read(uint8_t address, int numberOfRegisters)
{
  /// tell accelerometer which address we would like to access
  Wire.beginTransmission(DEVICE);
  Wire.write(address);
  Wire.endTransmission();

  /// get register contents
  Wire.beginTransmission(DEVICE);
  Wire.requestFrom((int)DEVICE, numberOfRegisters);

  int i = 0;
  uint8_t buffer[numberOfRegisters];
  while(Wire.available())
  { 
    buffer[i] = Wire.read();  /// receive one uint8_t
    i++;
  }
  Wire.endTransmission();
  
  return buffer;
}

void Altitude::calibrate()
{
  uint8_t * buffer = this->read(AC1_MSB, 22);
  
  this->AC1 = (short) (buffer[0] << 8) | buffer[1];
  this->AC2 = (short) (buffer[2] << 8) | buffer[3];
  this->AC3 = (short) (buffer[4] << 8) | buffer[5];
  this->AC4 = (unsigned short) (buffer[6] << 8) | buffer[7];
  this->AC5 = (unsigned short) (buffer[8] << 8) | buffer[9];
  this->AC6 = (unsigned short) (buffer[10] << 8) | buffer[11];
  this->b1  = (short) (buffer[12] << 8) | buffer[13];
  this->B2  = (short) (buffer[14] << 8) | buffer[15];
  this->MB  = (short) (buffer[16] << 8) | buffer[17];
  this->MC  = (short) (buffer[18] << 8) | buffer[19];
  this->MD  = (short) (buffer[20] << 8) | buffer[21];
}

/// tell sensor that we want temperature output by writing temperature value to register 
void Altitude::requestTemperature()
{
  this->write(REQUEST_REGISTER, TEMPERATURE_REQUEST);  /// have to wait at least 4.5ms before we can read registers
}
void Altitude::requestPressure()
{
  this->write(REQUEST_REGISTER, PRESSURE_REQUEST);  /// have to wait at least 4.5ms before we can read registers
}

/// remember to wait at least 4.5ms before calling this function after calling requestTemperature()
long Altitude::readTemperature()
{
  uint8_t * buffer = this->read(REQUEST_MSB, 2);
  
  this->UT = (buffer[0] << 8) | buffer[1];  // MSB + LSB

  return this->UT;
}
/// wait 4.5ms
long Altitude::readPressure()
{
  uint8_t * buffer = this->read(REQUEST_MSB, 3);
  
  this->UP = (((unsigned long) buffer[0] << 16) + ((unsigned long) buffer[1] << 8) + (unsigned long) buffer[2]) >> (8-OSS);

  return this->UP;
}

float Altitude::trueTemperature()
{
  long x1 = ((this->UT - (long)this->AC6)*(long)this->AC5) >> 15;
  long x2 = ((long)this->MC << 11)/(x1 + (long)this->MD);
  this->B5 = x1 + x2;
  
  this->temperature = (float) ((this->B5 + 8) >> 4)/10 + temperatureCalibrate;
  
  return this->temperature;
}

long Altitude::truePressure()
{
  long B6 = this->B5 - 4000;
  long x1 = (this->B2 * ((B6*B6) >> 12)) >> 11;
  long x2 = (this->AC2 * B6) >> 11;
  long x3 = x1 + x2;
  
  long B3 = ((AC1*4 + x3) << this->OSS + 2) >> 4;
  x1 = (this->AC3 * B6) >> 13;
  x2 = (this->b1 * ((B6*B6) >> 12)) >> 16;
  x3 = (x1 + x2 + 2) / 4;
  
  unsigned long B4 = (this->AC4 * (unsigned long)(x3 + 32768)) >> 15;
  unsigned long B7 = ((unsigned long)this->UP - B3) * (50000 >> OSS);
  
  if (B7 < 0x80000000)
  {
    this->pressure = (B7*2)/B4;
  }
  else
  {
    this->pressure = (B7/B4)*2;
  }
  
  x1 = ((this->pressure >> 8) * (this->pressure >> 8) * 3038) >> 16;
  x2 = (-7357 * this->pressure) >> 16;
  
  this->pressure += (x1 + x2 + 3791) >> 4;
  
  return this->pressure;
}

float Altitude::calculate()
{
  this->altitude = (float)44330*(1 - pow((float)this->pressure/this->pressureOffset, 0.190263237)); /// 1/5.25587611 = 0.190263237
  return this->altitude;
}

float Altitude::get(long milliseconds)
{
  /// make sure the last register write was done for 4.5ms ago
  if (milliseconds-this->time > 4.5)
  {
    if (!this->state)  /// true indicate temperature reading
    {
      this->readTemperature();
      this->requestPressure();
    }
    else
    {
      this->readPressure();
      this->trueTemperature();
      this->truePressure();
      this->calculate();
      this->requestTemperature();
    }
    this->time = milliseconds;
    this->state = !this->state;
  }

  return this->altitude;
}

