/*
generalized class for obtaining gyro measurements
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com)
*/

#include <Arduino.h>

class Altitude
{
  private:
    static const float temperatureCalibrate = -4.0;
    static const short OSS = 0; /// oversampling setting 0...3   
    long pressureOffset;  /// Pascal at sea level
    long UT;                    /// always keep a UT storage such that we have a value while waiting 4.5ms
    long UP;                    /// same as UT but for pressure
    float temperature;
    long pressure;
    float altitude;
    long time;                  /// initial values such that first measurement is not too early
    boolean state;              /// used as a state in an emulated interrupt    
    
    /// calibration values
    short AC1;
    short AC2;
    short AC3;
    unsigned short AC4;
    unsigned short AC5;
    unsigned short AC6;
    short b1;           /// B1 is reserved by binary.h
    short B2;
    long  B5;
    short MB;
    short MC;
    short MD;
  
  private:
    ///
    ///                 VARIABLE          HEX                  TYPE  DEFAULT_VALUE
    ///
    static const uint8_t   DEVICE          = 0x77;
    
    /// calibration registers
    static const uint8_t   AC1_MSB         = 0xAA;
    static const uint8_t   AC1_LSB         = 0xAB;
    static const uint8_t   AC2_MSB         = 0xAC;
    static const uint8_t   AC2_LSB         = 0xAD;
    static const uint8_t   AC3_MSB         = 0xAE;
    static const uint8_t   AC3_LSB         = 0xAF;
    static const uint8_t   AC4_MSB         = 0xB0;
    static const uint8_t   AC4_LSB         = 0xB1;
    static const uint8_t   AC5_MSB         = 0xB2;
    static const uint8_t   AC5_LSB         = 0xB3;
    static const uint8_t   AC6_MSB         = 0xB4;
    static const uint8_t   AC6_LSB         = 0xB5;
    static const uint8_t   B1_MSB          = 0xB6;
    static const uint8_t   B1_LSB          = 0xB7;
    static const uint8_t   B2_MSB          = 0xB8;
    static const uint8_t   B2_LSB          = 0xB9;
    static const uint8_t   MB_MSB          = 0xBA;
    static const uint8_t   MB_LSB          = 0xBB;
    static const uint8_t   MC_MSB          = 0xBC;
    static const uint8_t   MC_LSB          = 0xBD;
    static const uint8_t   MD_MSB          = 0xBE;
    static const uint8_t   MD_LSB          = 0xBF;
    
    /// request register
    static const uint8_t  REQUEST_REGISTER    = 0xF4;
    static const uint8_t  REQUEST_MSB         = 0xF6;
    static const uint8_t  REQUEST_LSB         = 0xF7;
    
    /// request values
    static const uint8_t  TEMPERATURE_REQUEST = 0x2E;
    static const uint8_t  PRESSURE_REQUEST    = 0x34 + (OSS << 6);


  
  public:
    Altitude();
    void begin();
    void write(uint8_t address, uint8_t code);
    uint8_t * read(uint8_t address,int numberOfregisters = 1);
    void calibrate();
    void requestTemperature();
    void requestPressure();
    long readTemperature();
    long readPressure();
    float trueTemperature();
    long truePressure();
    float calculate();
    float get(long microseconds = 0);

};

