/*
generalized class for magnetic flux from poles using a HMC5883L 3-axis digital compass
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com)
*/

#include <Wire.h>
#include "Compass.h"
#include <Arduino.h>  /// used for min() and atan2()


void Compass::begin()
{ 
  Wire.begin(); // join i2c bus
  scale[0] = 1.0;
  scale[1] = 1.0;
  scale[2] = 1.0;
  
  //writeReg(REG_A, 0x70); // 8 samples averaged, 75Hz frequency, no artificial bias.
  //writeReg(REG_B, 0xA0);
  //writeReg(MODE_REGISTER, 0x00);
}


bool Compass::calibrate(int samples, int startGain)
{
  
  int gain;
  bool saturated;
  long int total[3] = {0,0,0};
  
  /// loop through all gains in case of saturation
  for (gain = startGain; gain < 8; gain++)
  {
    saturated = false;
        
    /// apply positive bias
    write(DEVICE, REG_A, 0x010 + 1);  /// comment this later
    for (int i=0;i<samples;i++)
    {
      enableSingle();
      getRaw();
      
      total[0] += this->axis[0];
      total[1] += this->axis[1];
      total[2] += this->axis[2];
      
      /// saturation detect
      if (-4096 >= min(axis[0],min(axis[1],axis[2])))
      {
          saturated = true;
          break;
      }
    }
    
    /// apply negative bias
    write(DEVICE, REG_A, 0x010 + 2);  /// comment this later
    for (int i=0;i<samples;i++)
    {
      enableSingle();
      getRaw();
      
      total[0] -= this->axis[0];
      total[1] -= this->axis[1];
      total[2] -= this->axis[2];
      
      /// saturation detect
      if (-4096 >= min(axis[0],min(axis[1],axis[2])))
      {
          saturated = true;
          break;
      }
    }
    
    if (!saturated)
    {
      break;
    }
  }
  
  long int low_limit  = SELF_TEST_POSITIVE_LOW *compassGainGauss[gain]*2*samples;
  long int high_limit = SELF_TEST_POSITIVE_HIGH*compassGainGauss[gain]*2*samples;
  if (!saturated && low_limit <= total[0] && (high_limit >= total[0]) && (low_limit <= total[1]) && (high_limit >= total[1]) && (low_limit <= total[2]) && (high_limit >= total[2]) )
  {   /*
          Successful calibration.
          Normalize the scale factors so all axis return the same range of values for the bias field.
          Factor of 2 is from summation of total of n_samples from both positive and negative bias.
      */
      scale[0] = (compassGainGauss[gain] * (SELF_TEST_X_AXIS*2))/(total[0]/samples);
      scale[1] = (compassGainGauss[gain] * (SELF_TEST_Y_AXIS*2))/(total[1]/samples);
      scale[2] = (compassGainGauss[gain] * (SELF_TEST_Z_AXIS*2))/(total[2]/samples);
  }
  /// remove bias
  write(DEVICE, REG_A, 0x010);

  return !saturated;
}

void Compass::write(uint8_t device, uint8_t address, uint8_t code)
{
  Wire.beginTransmission(device);
  Wire.write(address);
  Wire.write(code);
  Wire.endTransmission();
}


uint8_t * Compass::read(uint8_t device, uint8_t address, int numberOfRegisters)
{
  /// tell accelerometer which address we would like to access
  Wire.beginTransmission(device);
  Wire.write(address);
  Wire.endTransmission();

  /// get register contents
  Wire.beginTransmission(device);
  Wire.requestFrom((int)device, numberOfRegisters);

  int i = 0;
  while(Wire.available())
  { 
    this->buffer[i] = Wire.read();  /// receive one uint8_t
    i++;
  }
  Wire.endTransmission();
  
  return this->buffer;
}

void Compass::getRaw()
{
  uint8_t * b = read(DEVICE, DATA_OUTPUT_X_MSB, 6);  /// read from DATA_OUTPUT_X_MSB and the next registersOfcompassData registers
  
  axis[0] = (b[0] << 8) | b[1];  /// X
  axis[2] = (b[2] << 8) | b[3];  /// Z  -  note the order
  axis[1] = (b[4] << 8) | b[5];  /// Y
}

void Compass::get()
{
  getRaw();
  this->axis[0] /= scale[0];
  this->axis[1] /= scale[1];
  this->axis[2] /= scale[2];
}


float Compass::x()
{
  return this->axis[0];
}
float Compass::y()
{
  return this->axis[1];
}
float Compass::z()
{
  return this->axis[2];
}

void Compass::setGain(int index)
{
  uint8_t value = compassGainAddress[1];

  if (index >= 0 && index < 8)
  {
    value = compassGainAddress[index];
  }
  
  this->write(DEVICE, REG_B, value);
}

/// put compass in continuous measuring mode
void Compass::enableContinuous()
{
  this->write(DEVICE, MODE_REGISTER, CONTINUOUS);
}

/// put compass in single mode
void Compass::enableSingle()
{
  this->write(DEVICE, MODE_REGISTER, SINGLE);
}

/// put compass in idle mode
void Compass::enableIdle()
{
  this->write(DEVICE, MODE_REGISTER, IDLE);
}


float Compass::roll(float declination)  /// phi
{
  float angle = ( atan2(y(),z()) + declination)*RAD_TO_DEG;
  if (angle < 0.0)
  {
    angle += 360;
  }
  return angle;  /// Use heading*M_PI/180 to obtain radians
}

float Compass::pitch(float declination)  /// theta
{
  float angle = ( atan2(-x(),z()) + declination)*RAD_TO_DEG;
  if (angle < 0.0)
  {
    angle += 360;
  }
  return angle;
}


float Compass::heading(float declination)  /// psi
{
  float heading = ( atan2(-y(), x()) + declination)*RAD_TO_DEG;
  
  /// correct for when signs are reversed.
  if (heading < 0.0)
  {
    heading += 360;
  }
  /// check for wrap due to addition of declination.
  else if (heading > 360.0)
  {
    heading -= 360;
  }

  return heading;
}
