/*
generalized class for obtaining compass measurements
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com)
*/



static const int   compassGainGauss[8]     = { 1370, 1090, 820,   660,  440,  390,  330,  230 };
static const uint8_t  compassGainAddress[8]   = { 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07 };  /// +/- 0.88, 1.3, 1.9, 2.5, 4.0, 4.7, 5.6, 8.1 Gau

class Compass
{
  private:
    
    static const uint8_t  DEVICE         = 0x1E;    // address to the device
    
    /// The table below shows nominal gain settings. Use the “Gain” column to convert counts to Gauss. Choose a lower gain value (higher GN#) when total field strength causes overflow in one of the data output registers (saturation).
    static const uint8_t  REG_A          = 0x00;    /// read/write

    
    /// rate
    static const uint8_t  REG_B          = 0x01;    /// read/write - Configuration Register B for configurating output frequency
    static const uint8_t  HZ_075         = 0x00;    /// 0.7 5Hz
    static const uint8_t  HZ_1_5         = 0x01;    /// 1.5 Hz
    static const uint8_t  HZ_3           = 0x02;    /// 3 Hz
    static const uint8_t  HZ_7_5         = 0x03;    /// 7.5 Hz
    static const uint8_t  HZ_15          = 0x04;    /// 15 Hz (default)
    static const uint8_t  HZ_30          = 0x05;    /// 30 Hz
    static const uint8_t  HZ_75          = 0x06;    /// 75 Hz

    static const uint8_t  MODE_REGISTER  = 0x02;    /// read/write
    static const uint8_t  CONTINUOUS     = 0x00;    /// Continuous-Measurement Mode. In continuous-measurement mode, the device continuously performs measurements and places the result in the data register.
                                                 /// RDY goes high when new data is placed in all three registers. After a power-on or a write to the mode or configuration register,
                                                 /// the first measurement set is available from all three data output registers after a period of 2/fDO and subsequent measurements are available at a frequency of fDO, where fDO is the frequency of data output.
    static const uint8_t  SINGLE         = 0x01;    /// Single-Measurement Mode (Default). When single-measurement mode is selected, device performs a single measurement, sets RDY high and returned to idle mode. Mode register returns to idle mode
                                                 /// bit values. The measurement remains in the data output register and RDY remains high until the data output register is read or another measurement is performed.
    static const uint8_t  IDLE           = 0x02;    /// Idle Mode. Device is placed in idle mode.

    /// the order X, Z, Y is correct in this matter
    static const uint8_t  DATA_OUTPUT_X_MSB          = 0x03;  /// read
    static const uint8_t  DATA_OUTPUT_X_LSB          = 0x04;  /// read
    static const uint8_t  DATA_OUTPUT_Z_MSB          = 0x05;  /// read
    static const uint8_t  DATA_OUTPUT_Z_LSB          = 0x06;  /// read
    static const uint8_t  DATA_OUTPUT_Y_MSB          = 0x07;  /// read
    static const uint8_t  DATA_OUTPUT_Y_LSB          = 0x08;  /// read
    static const uint8_t  STATUS_REGISTER            = 0x09;  /// read
    static const uint8_t  IDENTIFICATION_REGISTER_A  = 0x0A;  /// read
    static const uint8_t  IDENTIFICATION_REGISTER_B  = 0x0B;  /// read

    
    #if defined(__MK20DX128__)
    static constexpr float SELF_TEST_X_AXIS           = 1.16;
    static constexpr float SELF_TEST_Y_AXIS           = 1.16;
    static constexpr float SELF_TEST_Z_AXIS           = 1.08;
    static constexpr float SELF_TEST_POSITIVE_LOW     = (243/390);  /// 243 is given at gain=5, thus 390.
    static constexpr float SELF_TEST_POSITIVE_HIGH    = (575/390);
    static constexpr float SELF_TEST_NEGATIVE_LOW     = -(243/390);
    static constexpr float SELF_TEST_NEGATIVE_HIGH    = -(575/390);
    #else
    static const float SELF_TEST_X_AXIS           = 1.16;
    static const float SELF_TEST_Y_AXIS           = 1.16;
    static const float SELF_TEST_Z_AXIS           = 1.08;
    static const float SELF_TEST_POSITIVE_LOW     = (243/390);  /// 243 is given at gain=5, thus 390.
    static const float SELF_TEST_POSITIVE_HIGH    = (575/390);
    static const float SELF_TEST_NEGATIVE_LOW     = -(243/390);
    static const float SELF_TEST_NEGATIVE_HIGH    = -(575/390);
    #endif
    
    
  private:
    uint8_t buffer[6];
    int16_t scale[3];
    int16_t axis[3];

  
  public:
    void begin();
    bool calibrate(int samples, int startGain = 1);
    void write(uint8_t device, uint8_t address, uint8_t code);
    uint8_t * read(uint8_t device, uint8_t address, int numberOfregisters = 1);
    void getRaw();
    void get();
    float x();
    float y();
    float z();
    
    void setGain(int index = 1);
    void enableContinuous();
    void enableSingle();
    void enableIdle();
    float roll(float declination = 0.0372);
    float pitch(float declination = 0.0372);
    float heading(float declination = 0.0372);  /// default declination of Trondheim, Norway
};
