/*
generalized class for obtaining inertia from L3G4200D
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com)
*/

#include <Wire.h>
#include "Gyro.h"



bool Gyro::begin()
{
  bias[0] = 0;
  bias[1] = 0;
  bias[2] = 0;
  Wire.begin();                // join i2c bus
  
  if (this->read(DEVICE, WHO_AM_I)[0] != 0xD3)
  {
    return false;
  }

  write(DEVICE, CTRL_REG1, 0b10001111);  /// Power on, enable all axes and put in 400 Hz mode
  write(DEVICE, CTRL_REG2, 0b00000000);  /// Disable all filters
  write(DEVICE, CTRL_REG3, 0b00001000);
  write(DEVICE, CTRL_REG4, B00110000);   /// Set block mode, LSB, 2000 dps, disable self test and 4-wire I2C interface
  write(DEVICE, CTRL_REG5, 0b00000000);
  
  //write(FIFO_CTRL_REG, 0b00100000);
  
  return true;
}

bool Gyro::calibrate(bool calibrateXaxis, bool calibrateYaxis, bool calibrateZaxis)
{
  int16_t sum[3]     = {0,0,0};
  int16_t prevVal[3] = {0,0,0};
  
  for(int i=0; i<128;i++)
  {
    getRaw();

    /// Do not accept large changes
    if (abs(prevVal[0]-raw[0]) > 10 || abs(prevVal[1]-raw[1]) > 10 || abs(prevVal[2]-raw[2]) > 10)
    {
      i--;
    }
    else
    {
      sum[0] += raw[0];
      sum[1] += raw[1];
      sum[2] += raw[2];
    }
    prevVal[0] = raw[0];
    prevVal[1] = raw[1];
    prevVal[2] = raw[2];
    delay(5);  /// must wait due to fast loop, sleep is permitted here since this is an initation routine
  }
  if (calibrateXaxis) bias[0] = sum[0] / 128.0;
  if (calibrateYaxis) bias[1] = sum[1] / 128.0;
  if (calibrateZaxis) bias[2] = sum[2] / 128.0;
  
  return true;
}

void Gyro::write(uint8_t device, uint8_t address, uint8_t code)
{
  Wire.beginTransmission(device);
  Wire.write(address);
  Wire.write(code);
  Wire.endTransmission();
}


uint8_t * Gyro::read(uint8_t device, uint8_t address,int numberOfRegisters)
{
  /// tell accelerometer which address we would like to access
  Wire.beginTransmission(device);
  Wire.write(address);
  Wire.endTransmission();

  /// get register contents
  Wire.beginTransmission(device);
  Wire.requestFrom((int)device, numberOfRegisters);

  int i = 0;
  while(Wire.available())
  { 
    this->buffer[i] = Wire.read();  /// receive one uint8_t
    i++;
  }
  Wire.endTransmission();
  
  return this->buffer;
}

void Gyro::getRaw()
{
  uint8_t * b = read(DEVICE, OUT_X_L | (1 << 7), 6);
  
  this->raw[0] = (b[1] << 8) | b[0];
  this->raw[1] = (b[3] << 8) | b[2];
  this->raw[2] = (b[5] << 8) | b[4];
  
}


void Gyro::get()
{
  getRaw();
   
  /// Do not accept high values for Z due to an unknown disturbance. This is a very strange behaviour.
  if (abs(raw[2] - bias[2]) > 29750)
  {
    return;
  }
  
  /// FIFO stack for INS purposes
  xAxis[2] = xAxis[1];
  xAxis[1] = xAxis[0];
  xAxis[0] = raw[0] - bias[0];
  yAxis[2] = yAxis[1];
  yAxis[1] = yAxis[0];
  yAxis[0] = raw[1] - bias[1];
  zAxis[2] = zAxis[1];
  zAxis[1] = zAxis[0];
  zAxis[0] = raw[2] - bias[2];
}

/// meh
void Gyro::getAverage()
{
  
  getRaw();
  
  /// Do not accept high values for Z due to an unknown disturbance
  if (abs(raw[2] - bias[2]) > 29750)
  {
    return;
  }
  
  long total[3] = {0,0,0};

  /// FIFO stack
  for (int i=9;i>0;i--)
  {
    xAvg[i]   = xAvg[i-1];
    total[0] += xAvg[i-1];
    yAvg[i]   = yAvg[i-1];
    total[1] += yAvg[i-1];
    zAvg[i]   = zAvg[i-1];
    total[2] += zAvg[i-1];
  }
  xAvg[0] = raw[0] - bias[0];
  yAvg[0] = raw[1] - bias[1];
  zAvg[0] = raw[2] - bias[2];
  
  total[0] += xAvg[0];
  total[1] += yAvg[0];
  total[2] += zAvg[0];
  
  /// FIFO stack for INS purposes
  xAxis[2] = xAxis[1];
  xAxis[1] = xAxis[0];
  xAxis[0] = total[0]/10;
  yAxis[2] = yAxis[1];
  yAxis[1] = yAxis[0];
  yAxis[0] = total[1]/10;
  zAxis[2] = zAxis[1];
  zAxis[1] = zAxis[0];
  zAxis[0] = total[2]/10;
}

 

float Gyro::axis(int axis,int index)
{
  /// 8.75 is a gain value from the datasheet according to 250 degree per second
  if (index > 2)
  {
    index = 0;
  }
  if (axis == 1)
  {
    return yAxis[index]*this->_gain;  /// found by testing, but it is locatedish in datasheet
  }
  else if (axis == 2)
  {
    return -zAxis[index]*this->_gain;
  }
  else
  {
    return xAxis[index]*this->_gain;
  }
}

float Gyro::x(int index)
{
  return axis(0,index);
}
float Gyro::y(int index)
{
  return axis(1,index);
}
float Gyro::z(int index)
{
  return axis(2,index);
}

float Gyro::heading()
{
  return atan2(-yAxis[0],xAxis[0]);  /// not sure if this is right 
}

float Gyro::latitude()
{
  return atan2(-zAxis[0], sqrt( xAxis[0]*xAxis[0] + yAxis[0]*yAxis[0] ));  /// this is the formula, but not sure if it will work
}

int Gyro::temperature()
{
  /// in celsius, +273.15 to get in Kelvin
  return (int) read(DEVICE, this->OUT_TEMP,1)[0];
}
