/*
generalized class for obtaining inertia
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during winter 2013
*/

#include <stdlib.h>	/// abs()
#include <Arduino.h>	/// Use AVRs delay() for sleeping
#include "Gyro.h"



template <typename Driver> bool Gyro<Driver>::initialize(float gain)
{
  this->_gain = gain;

  if (_driver.initialize())
  {
  	return true;
  }
  
  return false;
}


/// Set bias for each axis, used by calibrate()
template <typename Driver> void Gyro<Driver>::bias(float xBias, float yBias, float zBias)
{
  _bias[0] = xBias;
  _bias[1] = yBias;
  _bias[2] = zBias;
}


/// Calibrate axes
template <typename Driver> bool Gyro<Driver>::calibrate(bool calibrateXaxis, bool calibrateYaxis, bool calibrateZaxis, unsigned threshold, unsigned int samples)
{
  int16_t sum[3]     = {0,0,0};
  int16_t prevVal[3] = {0,0,0};
  int16_t * raw;
  int moveCounter 	= 0;
  
  for(int i=0; i<samples;i++)
  {
    raw = _driver.measure();

    /// Detect and count movement
    if (abs(prevVal[0] - raw[0]) > threshold || abs(prevVal[1] - raw[1]) > threshold || abs(prevVal[2] - raw[2]) > threshold)
    {
      moveCounter++;
    }
    else
    {
      sum[0] += raw[0];
      sum[1] += raw[1];
      sum[2] += raw[2];
    }
    prevVal[0] = raw[0];
    prevVal[1] = raw[1];
    prevVal[2] = raw[2];
    delay(5);  /// must wait due to fast loop, sleep is permitted here since this is an initation routine
    
    /// if 10% movement has occurred, discontinue calibration
    if ((float)moveCounter/samples >= 0.1)
    {
    	return false;
    }
  }
  if (calibrateXaxis) _bias[0] = sum[0] / (float) samples;
  if (calibrateYaxis) _bias[1] = sum[1] / (float) samples;
  if (calibrateZaxis) _bias[2] = sum[2] / (float) samples;
  
  return true;
}


/// Measure movements
template <typename Driver> void Gyro<Driver>::measure()
{
  int16_t * raw = _driver.measure();  /// 3 axis, 3 values
  
  /// FIFO stack
  _xAxis[2] = _xAxis[1];
  _xAxis[1] = _xAxis[0];
  _xAxis[0] = raw[0] - _bias[0];
  _yAxis[2] = _yAxis[1];
  _yAxis[1] = _yAxis[0];
  _yAxis[0] = raw[1] - _bias[1];
  _zAxis[2] = _zAxis[1];
  _zAxis[1] = _zAxis[0];
  _zAxis[0] = raw[2] - _bias[2];
}


/// Average measurements. Do "num" averaging.
template <typename Driver> void Gyro<Driver>::getAverage(int num)
{
	if (num == 0)
	{
		return;
	}
  int16_t * raw = _driver.measure();
  
  long total[3] = {0,0,0};

  /// FIFO stack
  for (int i=num;i>0;i--)
  {
    _xAvg[i]  = _xAvg[i-1];
    total[0] += _xAvg[i-1];
    _yAvg[i]  = _yAvg[i-1];
    total[1] += _yAvg[i-1];
    _zAvg[i]  = _zAvg[i-1];
    total[2] += _zAvg[i-1];
  }
  _xAvg[0] = raw[0] - _bias[0];
  _yAvg[0] = raw[1] - _bias[1];
  _zAvg[0] = raw[2] - _bias[2];
  
  total[0] += _xAvg[0];
  total[1] += _yAvg[0];
  total[2] += _zAvg[0];
  
  /// FIFO stack
  _xAxis[2] = _xAxis[1];
  _xAxis[1] = _xAxis[0];
  _xAxis[0] = total[0]/num;
  _yAxis[2] = _yAxis[1];
  _yAxis[1] = _yAxis[0];
  _yAxis[0] = total[1]/num;
  _zAxis[2] = _zAxis[1];
  _zAxis[1] = _zAxis[0];
  _zAxis[0] = total[2]/num;
}



template <typename Driver> float Gyro<Driver>::axis(int axis,int index)
{
  /// do not get exceed array size
  if (index > 2)
  {
    index = 0;
  }
  
  if (axis == 1)
  {
    return _yAxis[index]*this->_gain;
  }
  else if (axis == 2)
  {
    return _zAxis[index]*this->_gain;
  }
  else
  {
    return _xAxis[index]*this->_gain;
  }
}


/// Allow setting of axis values such that the altered values are included in the FIFO stack
template <typename Driver> void Gyro<Driver>::setAxisValue(int axis, int index, double value)
{
	/// do not get exceed array size
	if (index > 2)
	{
		index = 0;
	}

	if (axis == 1)
	{
		_yAxis[index] = value;
	}
	else if (axis == 2)
	{
		_zAxis[index] = value;
	}
	else
	{
		_xAxis[index] = value;
	}
}


template <typename Driver> float Gyro<Driver>::x(int index)
{
  return axis(0,index);
}
template <typename Driver> float Gyro<Driver>::y(int index)
{
  return axis(1,index);
}
template <typename Driver> float Gyro<Driver>::z(int index)
{
  return axis(2,index);
}



template <typename Driver> float Gyro<Driver>::heading()
{
  return atan2(-y(),x());  /// not sure if this is right 
}



template <typename Driver> float Gyro<Driver>::latitude()
{
  return atan2(-z(), sqrt( x()*x() + y()*y() ));  /// this is the formula, but not sure if it will work
}


/// Some chips has temperature sensors embedded for better performance, lets support this
template <typename Driver> int Gyro<Driver>::temperature()
{
  return _driver.temperature();
}
