/*
Generalized class for reading magnetic flux from poles using a arbitrarily 3-axis magnetometer
Written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during winter 2013
*/

#include "Compass.h"


template <typename Driver> void Compass<Driver>::initialize()
{ 
  _driver.initialize();
}



template <typename Driver> void Compass<Driver>::measure()
{
  int16_t * raw = _driver.measure();
  
  /// FIFO stack
  _xAxis[2] = _xAxis[1];
  _xAxis[1] = _xAxis[0];
  _xAxis[0] = raw[0];
  _yAxis[2] = _yAxis[1];
  _yAxis[1] = _yAxis[0];
  _yAxis[0] = raw[1];
  _zAxis[2] = _zAxis[1];
  _zAxis[1] = _zAxis[0];
  _zAxis[0] = raw[2];
}


/// Get axis value
template <typename Driver> float Compass<Driver>::axis(int axis,int index)
{
  /// do not get exceed array size
  if (index > 2)
  {
    index = 0;
  }
  
  if (axis == 1)
  {
    return _yAxis[index];
  }
  else if (axis == 2)
  {
    return _zAxis[index];
  }
  else
  {
    return _xAxis[index];
  }
}



template <typename Driver> float Compass<Driver>::x(int index)
{
  return axis(0,index);
}
template <typename Driver> float Compass<Driver>::y(int index)
{
  return axis(1,index);
}
template <typename Driver> float Compass<Driver>::z(int index)
{
  return axis(2,index);
}


template <typename Driver> void Compass<Driver>::setGain(int index)
{
	_driver.setGain(index);
}


/// put compass in continuous measuring mode
template <typename Driver> void Compass<Driver>::continuous()
{
  _driver.continuous();
}


/// put compass in single mode
template <typename Driver> void Compass<Driver>::periodic()
{
	_driver.periodic();
}


/// put compass in idle mode
template <typename Driver> void Compass<Driver>::idle()
{
	_driver.idle();
}



template <typename Driver> float Compass<Driver>::roll(float declination)  /// phi
{
  float angle = ( atan2(y(),z()) + declination)*(180/M_PI);	/// convert to degrees
  if (angle < 0.0)
  {
    angle += 360;
  }
  
  return angle;  /// Use heading*M_PI/180 to obtain radians
}



template <typename Driver> float Compass<Driver>::pitch(float declination)  /// theta
{
  float angle = ( atan2(x(),z()) + declination)*(180/M_PI);	/// convert to degrees
  if (angle < 0.0)
  {
    angle += 360;
  }
  
  return angle;
}



template <typename Driver> float Compass<Driver>::heading(float declination)  /// psi
{
  float heading = ( atan2(y(), x()) + declination)*(180/M_PI);	/// convert to degrees
  
  /// correct for when signs are reversed.
  if (heading < 0.0)
  {
    heading += 360;
  }
  /// check for wrap due to addition of declination.
  else if (heading > 360.0)
  {
    heading -= 360;
  }

  return heading;
}



template <typename Driver> bool Compass<Driver>::selfTest()
{
	if (_driver.selfTest())
	{
		return true;
	}
	return false;
}



template <typename Driver> bool Compass<Driver>::calibrate()
{
	/// not implemented
  return true;
}


/// Allow setting of axis values such that the altered values are included in the FIFO stack
template <typename Driver> void Compass<Driver>::setAxisValue(int axis, int index, double value)
{
	/// do not get exceed array size
	if (index > 2)
	{
		index = 0;
	}

	if (axis == 1)
	{
		_yAxis[index] = value;
	}
	else if (axis == 2)
	{
		_zAxis[index] = value;
	}
	else
	{
		_xAxis[index] = value;
	}
}