/*
This file handles communications back and forth with javascript inside Qwebview
Written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during winter 2013
*/

#include <QDebug>
#include "Vehicle.hpp"
#include "Waypoint.hpp"
#include "JsBridge.hpp"
#include "mainwindow.hpp"
#include <QStringList>



JsBridge::JsBridge(QObject *parent) : QObject(parent)
{
    _IDENTITY_PREFIX = "F";
    _parent     = parent;
    _vehicle    = new Vehicle(this);
    _waypoint   = new Waypoint(this);
}



JsBridge::~JsBridge()
{
    delete _parent;
}


/// Incomming serial data is passed to this handler
void JsBridge::serialDataHandle(QString data)
{
    /// Look for new vehicle positions
    QRegExp rx("^"+_IDENTITY_PREFIX+"([^ ]*) P ([^ ]*) ([^ ]*) (.*)$");
    if (rx.indexIn(data) != -1)
    {
        QStringList list    = rx.capturedTexts();
        QString id          = _IDENTITY_PREFIX+list.at(1);
        float latitude      = list.at(2).toFloat();
        float longitude     = list.at(3).toFloat();
        float altitude      = list.at(4).toFloat();

        /// Add vehicle if not existing
        if (!_vehicle->exists(id))
        {
            _vehicle->add(id, latitude, longitude, altitude);
        }
        else
        {
            _vehicle->move(id, latitude, longitude, altitude);
        }
    }
}



void JsBridge::waypoint(double latitude, double longitude, double altitude)
{
    _waypoint->add(latitude,longitude,altitude);
}


/// All waypoints are given, upload
void JsBridge::upload()
{
    MainWindow * parent = qobject_cast<MainWindow*>(_parent);
    if (_vehicle->list().length() == 0)
    {
        parent->setStatus("No drone is present");
    }
    else
    {
        MainWindow * parent = qobject_cast<MainWindow*>(_parent);
        QStringList vehicles = _vehicle->list();
        QList<Waypoint::geodetic> waypoints = _waypoint->coordinates();
        QString data;

        /// Loop through all connected vehicles..
        for (int i=0; i < vehicles.length(); i++)
        {
            /// ..and send all waypoints
            for (int j=0; j < waypoints.length(); j++)
            {
                /// Format serial data: ID WPA LAT LON ALT\n
                data = vehicles.at(i) + " WPA "+QString::number(waypoints.at(j).latitude,'f',8)+" "+QString::number(waypoints.at(j).longitude,'f',8)+" "+QString::number(waypoints.at(j).altitude,'f',8)+"\n";    /// WPA = WayPoint Add = create new way point

                /// Write data to serial. Send 8 byte each 100ms
                parent->_serial->write(data, 8, 100);
            }
        }

        /// Remove all waypoints till next time
        _waypoint->clear();

        parent->setStatus("Upload complete");
    }
}



void JsBridge::setStatus(QString message)
{
    MainWindow * parent = qobject_cast<MainWindow*>(_parent);
    parent->setStatus(message);
}


/// center map around this point
void JsBridge::center(double latitude, double longitude)
{
    QString js = "map.setCenter(new google.maps.LatLng(" + QString::number(latitude, 'f', 8) + ", " + QString::number(longitude, 'f', 8) + "));";

    MainWindow * parent = qobject_cast<MainWindow*>(_parent);
    parent->evaluateJs(js);
}


/// set zoom level of map
void JsBridge::zoom(int zoom)
{
    QString js = "map.zoom(" + QString::number(zoom, 'g', 0) + ");";

    MainWindow * parent = qobject_cast<MainWindow*>(_parent);
    parent->evaluateJs(js);
}


/// Make vehicle return to origin
void JsBridge::returnHome(QString id)
{
    MainWindow * parent = qobject_cast<MainWindow*>(_parent);
    parent->_serial->write(id + " RETURN\n");
}


/// Send JS to QWebFrame
void JsBridge::evaluateJs(QString js)
{
    MainWindow * parent = qobject_cast<MainWindow*>(_parent);
    parent->evaluateJs(js);
}


/// Clear map, vehicles and way-points
void JsBridge::clear()
{
    _vehicle->clear();
    _waypoint->clear();
}
