/*
Handle serial communication with USB device
Written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during winter 2013
*/


#include "mainwindow.hpp"
#include "Serial.hpp"
#include "Misc.hpp"
#include <QtSerialPort/QSerialPort>
#include <QtSerialPort/QSerialPortInfo>
#include <QStringList>
#include <QThread>



Serial::Serial(QObject *parent, JsBridge *jsb) : QObject(parent)
{
    _parent = parent;
    _jsBridge = jsb;
    _serial = new QSerialPort(parent);

    /// When data is ready, QSerialPort::readyRead() returns true and Serial::readData() is called
    _parent->connect(_serial, SIGNAL(readyRead()), this, SLOT(readData()));
}



Serial::~Serial()
{
    delete _parent;
    delete _serial;
}



bool Serial::connect(QString port, qint64 baudRate, QSerialPort::DataBits dataBits, QSerialPort::Parity parity, QSerialPort::StopBits stopBits, QSerialPort::FlowControl flowControl)
{
    /// find port automatically
    if (port == "")
    {
        foreach (const QSerialPortInfo &info, QSerialPortInfo::availablePorts())
        {
            port = info.portName();
            break;
        }
    }

    if (port == "")
    {
        return false;
    }

    _serial->setPortName(port);
    if (!_serial->open(QIODevice::ReadWrite))
    {
        return false;
    }
    if (!_serial->setBaudRate( baudRate ))
    {
        return false;
    }
    if (!_serial->setDataBits( dataBits ))
    {
        return false;
    }
    if (!_serial->setParity( parity ))
    {
        return false;
    }
    if (!_serial->setStopBits( stopBits ))
    {
        return false;
    }
    if (!_serial->setFlowControl( flowControl ))
    {
        return false;
    }

    _port = port;
    return true;
}


/// Write string to serial port
void Serial::write(QString data, int bufferSize, long milliseconds)
{
    /// Send bytes then wait milliseconds
    for (int i=0; i < data.length(); i += bufferSize)
    {
        /// Split into chunks
        QByteArray qba;
        qba.append(data.mid(i,bufferSize));

        /// Send chunks
        _serial->write(qba);

        /// Wait
        Misc::usleep(milliseconds);
    }
}


/// Read incomming data from serial port
void Serial::readData()
{
    QString raw  = QString(_serial->readAll());
    QString data = bufferUntil(raw, QString("\n"));

    if (data == "")
    {
        return;
    }

    /// send data to parent function
    _jsBridge->serialDataHandle(data);
}


/// Return a string between each new line
QString Serial::bufferUntil(QString data, QString delimiter)
{
    _buffer.append(data);

    QStringList list = _buffer.split(delimiter);
    QString returnValue = "";

    for (int i=0; i < list.length()-1; i++)
    {
        returnValue.append(list.at(i));
    }

    _buffer = list.at(list.length()-1);

    return returnValue;
}



QString Serial::port()
{
    return _port;
}



QString Serial::error()
{
    return _serial->errorString();
}
