/*
Find red colors from a connected USB camera
written by Vegard Hammerseth <vegard@hammerseth.com> (http://vegard.hammerseth.com) during spring 2013

compile with: g++ -ggdb `pkg-config --cflags opencv` -o find_red_live find_red_live.cpp `pkg-config --libs opencv`
*/


//#define GUI 1	/// Use this when debugging on PC


char _DEVICE[]  = "/dev/ttyUSB0";
char _DEVICE2[] = "/dev/ttyUSB1";
int _BAUD		 = 115200;

int _thresholdBlue = 50;
int _thresholdGreen = 50;
int _thresholdRed = 80;
int _dilate = 0;
int _erode = 3;



/// Includes
#include <sstream>
#include <string>
#include "Serial.cpp"
#include <stdio.h>
#include <ctime>		/// time()
#include <opencv2/opencv.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/stitching/stitcher.hpp>

using namespace cv;


/// Find red from an image and return a binary image
void findRed(cv::Mat& img, cv::Mat& out, int thresholdBlue = 100, int thresholdGreen = 100, int thresholdRed = 150)
{
	img.copyTo(out);
	
	inRange(img, Scalar(0, 0, thresholdRed), Scalar(thresholdBlue, thresholdGreen, 255), out);
	out = out < 128;
}


/// Get the largest contour in an image
double findLargestContour(cv::Mat& img, cv::Mat& out)
{
	cv::Mat analyze;
	img.copyTo(analyze);
	
	vector<Vec4i> hierarchy;
	vector<vector<Point> > contours;
	cv::findContours( analyze, contours, hierarchy, CV_RETR_TREE, CV_CHAIN_APPROX_SIMPLE, Point(0, 0) );

	double area;
	double maxArea = 0;;
	int maxIndex = 0;
	
	for( int i = 1; i < contours.size(); i++ )	/// Start at 1 to skip entire image contour
	{
		area = contourArea(contours[i]);
		if (area > maxArea)
		{
			maxArea = area;
			maxIndex = i;
		}
	}
	
	if (maxIndex != 0)
	{
		drawContours( out, contours, maxIndex, Scalar(0,255,0), 1, 8, hierarchy, 0, Point() );
	}
	
	return maxArea;
}


int main(int argc, char** argv)
{
	printf("Syntax: %s [0/1] [blue] [green] [red] [dilate] [erode]\n", argv[0]);
	/// Allow parameters
	if (argc > 2)
	{
		_thresholdBlue = atoi(argv[2]);
	}
	if (argc > 3)
	{
		_thresholdGreen = atoi(argv[3]);
	}
	if (argc > 4)
	{
		_thresholdRed = atoi(argv[4]);
	}
	if (argc > 5)
	{
		_dilate = atoi(argv[5]);
	}
	if (argc > 6)
	{
		_erode = atoi(argv[6]);
	}
	
	
	int serialFd = serialport_init(_DEVICE,_BAUD);	// start serial communication
	bool toggle = false;
	while (serialFd == -1)	// wait till serial communication has started
	{
		toggle = !toggle;
		perror("failed to open port, waiting 5 seconds\n");
		sleep(5);
		serialFd = serialport_init( (!toggle ? _DEVICE : _DEVICE2), _BAUD);
	}
	printf("Connected to %s\n", (!toggle ? _DEVICE : _DEVICE2));
	sleep(5);
	//serialport_write(serialFd, "Booted");
	
	/// Initialize first connected video device
   cv::VideoCapture cam = cv::VideoCapture(-1);

   /// Set camera properties
   cam.set(CV_CAP_PROP_FRAME_WIDTH, 1280);
   cam.set(CV_CAP_PROP_FRAME_HEIGHT, 720);
	

	/// Image holders
	Mat frame;
	Mat modified;
	Mat result;
	clock_t start;

	#ifdef GUI	
	namedWindow("Image", CV_WINDOW_AUTOSIZE);
	#endif

	/// The c is only used when GUI is defined. Thus the loop will run forever when not defined.
	int c = 0;
	char b[1];
	
	std::string position = "";
	
	while (c != 27)
	{
		serialport_read_until(serialFd, buf, '$');	/// This is a tørrible, tørrible idea! 
		if (strlen(buf) > 3 && buf[0] == 'G' && buf[1] == 'P' && buf[2] == 'G' && buf[3] == 'G')
		{
			char * token = strtok(buf, "\n");
			position = string(token);
      	printf("%s\n",buf);
      }

		/// Read image from camera
		cam >> frame;
		
		//for (int tresh = 150; tresh >= _thresholdRed; tresh -= 25)
		//{	
			/// find red and receive binary image
			findRed(frame, modified, _thresholdBlue, _thresholdGreen, _thresholdRed);	/// three channels
			
			/// Remove unwanted objects and enhance the rest
			dilate( modified, modified, cv::getStructuringElement( cv::MORPH_RECT, cv::Size( 2*_dilate + 1, 2*_dilate+1 ), cv::Point( _dilate, _dilate ) ) );
			erode( modified, result, cv::getStructuringElement( cv::MORPH_RECT, cv::Size( 2*_erode + 1, 2*_erode+1 ), cv::Point( _erode, _erode ) ) );
		
			/// Found object
			if (findLargestContour(result, frame) > 200)
			{			
				/// Save image if parameter is given
				if (argc > 1 && atoi(argv[1]) == 1)
				{
					//putText(frame, position, cvPoint(5,20), FONT_HERSHEY_COMPLEX_SMALL, 0.8, cvScalar(0,255,0), 1, CV_AA);
					
					if (imwrite(position + ".jpeg",frame))
					{
						printf("Saved image\n");
					}
				}
			}
		//}
		
		/// Show captured frame if GUI is set
		#ifdef GUI
		imshow("Image", frame);
		c = cv::waitKey(25);	/// 67 = 15 FPS
		#endif
	}


	return 0;
}

